# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os
from pathlib import Path
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
)

import pytest

from playwright.async_api import (
    BrowserContext,
    BrowserType,
    Cookie,
    Error,
    Page,
    expect,
)
from tests.server import Server
from tests.utils import must

from .utils import Utils


@pytest.fixture()
async def launch_persistent(
    tmp_path: Path, launch_arguments: Dict, browser_type: BrowserType
) -> AsyncGenerator[Callable[..., Awaitable[Tuple[Page, BrowserContext]]], None]:
    context: Optional[BrowserContext] = None

    async def _launch(**options: Any) -> Tuple[Page, BrowserContext]:
        nonlocal context
        if context:
            raise ValueError("can only launch one persistent context")
        context = await browser_type.launch_persistent_context(
            str(tmp_path), **{**launch_arguments, **options}
        )
        assert context
        return (context.pages[0], context)

    yield _launch
    await must(context).close()


async def test_context_cookies_should_work(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
    default_same_site_cookie_value: str,
) -> None:
    (page, context) = await launch_persistent()
    await page.goto(server.EMPTY_PAGE)
    document_cookie = await page.evaluate(
        """() => {
    document.cookie = 'username=John Doe';
    return document.cookie;
  }"""
    )

    assert document_cookie == "username=John Doe"
    assert _filter_cookies(await page.context.cookies()) == [
        {
            "name": "username",
            "value": "John Doe",
            "domain": "localhost",
            "path": "/",
            "expires": -1,
            "httpOnly": False,
            "secure": False,
            "sameSite": default_same_site_cookie_value,
        }
    ]


async def test_context_add_cookies_should_work(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
    default_same_site_cookie_value: Literal["Lax", "None", "Strict"],
) -> None:
    (page, context) = await launch_persistent()
    await page.goto(server.EMPTY_PAGE)
    await page.context.add_cookies(
        [
            {
                "url": server.EMPTY_PAGE,
                "name": "username",
                "value": "John Doe",
                "sameSite": default_same_site_cookie_value,
            }
        ]
    )
    assert await page.evaluate("() => document.cookie") == "username=John Doe"
    assert _filter_cookies(await page.context.cookies()) == [
        {
            "name": "username",
            "value": "John Doe",
            "domain": "localhost",
            "path": "/",
            "expires": -1,
            "httpOnly": False,
            "secure": False,
            "sameSite": default_same_site_cookie_value,
        }
    ]


def _filter_cookies(cookies: Sequence[Cookie]) -> List[Cookie]:
    return list(
        filter(lambda cookie: not cookie["domain"].endswith("microsoft.com"), cookies)
    )


async def test_context_clear_cookies_should_work(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent()
    await page.goto(server.EMPTY_PAGE)
    await page.context.add_cookies(
        [
            {"url": server.EMPTY_PAGE, "name": "cookie1", "value": "1"},
            {"url": server.EMPTY_PAGE, "name": "cookie2", "value": "2"},
        ]
    )
    assert await page.evaluate("document.cookie") == "cookie1=1; cookie2=2"
    await page.context.clear_cookies()
    await page.reload()
    assert _filter_cookies(await page.context.cookies([])) == []
    assert await page.evaluate("document.cookie") == ""


async def test_should_not_block_third_party_cookies(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
    is_firefox: bool,
) -> None:
    (page, context) = await launch_persistent()
    await page.goto(server.EMPTY_PAGE)
    await page.evaluate(
        """src => {
    let fulfill;
    const promise = new Promise(x => fulfill = x);
    const iframe = document.createElement('iframe');
    document.body.appendChild(iframe);
    iframe.onload = fulfill;
    iframe.src = src;
    return promise;
  }""",
        server.CROSS_PROCESS_PREFIX + "/grid.html",
    )
    document_cookie = await page.frames[1].evaluate(
        """() => {
    document.cookie = 'username=John Doe';
    return document.cookie;
  }"""
    )

    await page.wait_for_timeout(2000)
    allows_third_party = is_firefox
    assert document_cookie == ("username=John Doe" if allows_third_party else "")
    cookies = await context.cookies(server.CROSS_PROCESS_PREFIX + "/grid.html")
    if allows_third_party:
        assert cookies == [
            {
                "domain": "127.0.0.1",
                "expires": -1,
                "httpOnly": False,
                "name": "username",
                "path": "/",
                "sameSite": "None",
                "secure": False,
                "value": "John Doe",
            }
        ]
    else:
        assert cookies == []


async def test_should_support_viewport_option(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
    utils: Utils,
) -> None:
    (page, context) = await launch_persistent(viewport={"width": 456, "height": 789})
    await utils.verify_viewport(page, 456, 789)
    page2 = await context.new_page()
    await utils.verify_viewport(page2, 456, 789)


async def test_should_support_device_scale_factor_option(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(device_scale_factor=3)
    assert await page.evaluate("window.devicePixelRatio") == 3


async def test_should_support_user_agent_option(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
    server: Server,
) -> None:
    (page, context) = await launch_persistent(user_agent="foobar")
    assert await page.evaluate("() => navigator.userAgent") == "foobar"
    [request, _] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        page.goto(server.EMPTY_PAGE),
    )
    assert request.getHeader("user-agent") == "foobar"


async def test_should_support_bypass_csp_option(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
    server: Server,
) -> None:
    (page, context) = await launch_persistent(bypass_csp=True)
    await page.goto(server.PREFIX + "/csp.html")
    await page.add_script_tag(content="window.__injected = 42;")
    assert await page.evaluate("() => window.__injected") == 42


async def test_should_support_javascript_enabled_option(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
    is_webkit: bool,
) -> None:
    (page, context) = await launch_persistent(java_script_enabled=False)
    await page.goto('data:text/html, <script>var something = "forbidden"</script>')
    with pytest.raises(Error) as exc:
        await page.evaluate("something")
    if is_webkit:
        assert "Can't find variable: something" in exc.value.message
    else:
        assert "something is not defined" in exc.value.message


async def test_should_support_http_credentials_option(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(
        http_credentials={"username": "user", "password": "pass"}
    )
    server.set_auth("/playground.html", "user", "pass")
    response = await page.goto(server.PREFIX + "/playground.html")
    assert response
    assert response.status == 200


async def test_should_support_offline_option(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(offline=True)
    with pytest.raises(Error):
        await page.goto(server.EMPTY_PAGE)


async def test_should_support_has_touch_option(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(has_touch=True)
    await page.goto(server.PREFIX + "/mobile.html")
    assert await page.evaluate('() => "ontouchstart" in window')


@pytest.mark.skip_browser("firefox")
async def test_should_work_in_persistent_context(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    # Firefox does not support mobile.
    (page, context) = await launch_persistent(
        viewport={"width": 320, "height": 480}, is_mobile=True
    )
    await page.goto(server.PREFIX + "/empty.html")
    assert await page.evaluate("() => window.innerWidth") == 980


async def test_should_support_color_scheme_option(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(color_scheme="dark")
    assert (
        await page.evaluate('() => matchMedia("(prefers-color-scheme: light)").matches')
        is False
    )
    assert await page.evaluate(
        '() => matchMedia("(prefers-color-scheme: dark)").matches'
    )


async def test_should_support_timezone_id_option(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(timezone_id="America/Jamaica")
    assert (
        await page.evaluate("() => new Date(1479579154987).toString()")
        == "Sat Nov 19 2016 13:12:34 GMT-0500 (Eastern Standard Time)"
    )


async def test_should_support_contrast_option(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, _) = await launch_persistent(contrast="more")
    assert await page.evaluate('() => matchMedia("(prefers-contrast: more)").matches')
    assert not await page.evaluate(
        '() => matchMedia("(prefers-contrast: no-preference)").matches'
    )


async def test_should_support_locale_option(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(locale="fr-FR")
    assert await page.evaluate("() => navigator.language") == "fr-FR"


async def test_should_support_geolocation_and_permission_option(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(
        geolocation={"longitude": 10, "latitude": 10}, permissions=["geolocation"]
    )
    await page.goto(server.EMPTY_PAGE)
    geolocation = await page.evaluate(
        """() => new Promise(resolve => navigator.geolocation.getCurrentPosition(position => {
    resolve({latitude: position.coords.latitude, longitude: position.coords.longitude});
  }))"""
    )
    assert geolocation == {"latitude": 10, "longitude": 10}


async def test_should_support_ignore_https_errors_option(
    https_server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(ignore_https_errors=True)
    response = await page.goto(https_server.EMPTY_PAGE)
    assert response
    assert response.ok


async def test_should_support_extra_http_headers_option(
    server: Server,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(extra_http_headers={"foo": "bar"})
    [request, _] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        page.goto(server.EMPTY_PAGE),
    )
    assert request.getHeader("foo") == "bar"


async def test_should_accept_user_data_dir(
    tmp_path: Path,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent()
    # Note: we need an open page to make sure its functional.
    assert len(os.listdir(tmp_path)) > 0
    await context.close()
    assert len(os.listdir(tmp_path)) > 0


async def test_should_restore_state_from_userDataDir(
    browser_type: BrowserType,
    launch_arguments: Dict,
    server: Server,
    tmp_path_factory: pytest.TempPathFactory,
) -> None:
    user_data_dir1 = tmp_path_factory.mktemp("test")
    browser_context = await browser_type.launch_persistent_context(
        user_data_dir1, **launch_arguments
    )
    page = await browser_context.new_page()
    await page.goto(server.EMPTY_PAGE)
    await page.evaluate('() => localStorage.hey = "hello"')
    await browser_context.close()

    browser_context2 = await browser_type.launch_persistent_context(
        user_data_dir1, **launch_arguments
    )
    page2 = await browser_context2.new_page()
    await page2.goto(server.EMPTY_PAGE)
    assert await page2.evaluate("() => localStorage.hey") == "hello"
    await browser_context2.close()

    user_data_dir2 = tmp_path_factory.mktemp("test")
    browser_context3 = await browser_type.launch_persistent_context(
        user_data_dir2, **launch_arguments
    )
    page3 = await browser_context3.new_page()
    await page3.goto(server.EMPTY_PAGE)
    assert await page3.evaluate("() => localStorage.hey") != "hello"
    await browser_context3.close()


async def test_should_have_default_url_when_launching_browser(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent()
    urls = list(map(lambda p: p.url, context.pages))
    assert urls == ["about:blank"]


@pytest.mark.skip_browser("firefox")
async def test_should_throw_if_page_argument_is_passed(
    browser_type: BrowserType, server: Server, tmp_path: Path, launch_arguments: Dict
) -> None:
    options = {**launch_arguments, "args": [server.EMPTY_PAGE]}
    with pytest.raises(Error) as exc:
        await browser_type.launch_persistent_context(tmp_path, **options)
    assert "can not specify page" in exc.value.message


async def test_should_fire_close_event_for_a_persistent_context(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent()
    fired_event: "asyncio.Future[bool]" = asyncio.Future()
    context.on("close", lambda _: fired_event.set_result(True))
    await context.close()
    await fired_event


async def test_should_support_reduced_motion(
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent(reduced_motion="reduce")
    assert await page.evaluate("matchMedia('(prefers-reduced-motion: reduce)').matches")


async def test_should_support_har_option(
    assetdir: Path,
    launch_persistent: "Callable[..., asyncio.Future[Tuple[Page, BrowserContext]]]",
) -> None:
    (page, context) = await launch_persistent()
    await page.route_from_har(har=assetdir / "har-fulfill.har")
    await page.goto("http://no.playwright/")
    assert await page.evaluate("window.value") == "foo"
    await expect(page.locator("body")).to_have_css("background-color", "rgb(255, 0, 0)")



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import threading
from pathlib import Path
from typing import Dict, Generator, Optional, cast

import OpenSSL.crypto
import OpenSSL.SSL
import pytest
from twisted.internet import reactor as _twisted_reactor
from twisted.internet import ssl
from twisted.internet.selectreactor import SelectReactor
from twisted.web import resource, server
from twisted.web.http import Request

from playwright.sync_api import Browser, BrowserType, Playwright, expect

reactor = cast(SelectReactor, _twisted_reactor)


@pytest.fixture(scope="function", autouse=True)
def _skip_webkit_darwin(browser_name: str) -> None:
    if browser_name == "webkit" and sys.platform == "darwin":
        pytest.skip("WebKit does not proxy localhost on macOS")


class HttpsResource(resource.Resource):
    serverCertificate: ssl.PrivateCertificate
    isLeaf = True

    def _verify_cert_chain(self, cert: Optional[OpenSSL.crypto.X509]) -> bool:
        if not cert:
            return False
        store = OpenSSL.crypto.X509Store()
        store.add_cert(self.serverCertificate.original)
        store_ctx = OpenSSL.crypto.X509StoreContext(store, cert)
        try:
            store_ctx.verify_certificate()
            return True
        except OpenSSL.crypto.X509StoreContextError:
            return False

    def render_GET(self, request: Request) -> bytes:
        tls_socket: OpenSSL.SSL.Connection = request.transport.getHandle()  # type: ignore
        cert = tls_socket.get_peer_certificate()
        parts = []

        if self._verify_cert_chain(cert):
            request.setResponseCode(200)
            parts.append(
                {
                    "key": "message",
                    "value": f"Hello {cert.get_subject().CN}, your certificate was issued by {cert.get_issuer().CN}!",  # type: ignore
                }
            )
        elif cert and cert.get_subject():
            request.setResponseCode(403)
            parts.append(
                {
                    "key": "message",
                    "value": f"Sorry {cert.get_subject().CN}, certificates from {cert.get_issuer().CN} are not welcome here.",
                }
            )
        else:
            request.setResponseCode(401)
            parts.append(
                {
                    "key": "message",
                    "value": "Sorry, but you need to provide a client certificate to continue.",
                }
            )
        return b"".join(
            [
                f'<div data-testid="{part["key"]}">{part["value"]}</div>'.encode()
                for part in parts
            ]
        )


@pytest.fixture(scope="session", autouse=True)
def _client_certificate_server(assetdir: Path) -> Generator[None, None, None]:
    certAuthCert = ssl.Certificate.loadPEM(
        (assetdir / "client-certificates/server/server_cert.pem").read_text()
    )
    serverCert = ssl.PrivateCertificate.loadPEM(
        (assetdir / "client-certificates/server/server_key.pem").read_text()
        + (assetdir / "client-certificates/server/server_cert.pem").read_text()
    )

    contextFactory = serverCert.options(certAuthCert)
    contextFactory.requireCertificate = False
    resource = HttpsResource()
    resource.serverCertificate = serverCert
    site = server.Site(resource)

    def _run() -> None:
        reactor.listenSSL(8000, site, contextFactory)

    thread = threading.Thread(target=_run)
    thread.start()
    yield
    thread.join()


def test_should_throw_with_untrusted_client_certs(
    playwright: Playwright, assetdir: Path
) -> None:
    serverURL = "https://localhost:8000/"
    request = playwright.request.new_context(
        # TODO: Remove this once we can pass a custom CA.
        ignore_https_errors=True,
        client_certificates=[
            {
                "origin": serverURL,
                "certPath": assetdir
                / "client-certificates/client/self-signed/cert.pem",
                "keyPath": assetdir / "client-certificates/client/self-signed/key.pem",
            }
        ],
    )
    with pytest.raises(Exception, match="alert unknown ca"):
        request.get(serverURL)
    request.dispose()


def test_should_work_with_new_context(browser: Browser, assetdir: Path) -> None:
    context = browser.new_context(
        # TODO: Remove this once we can pass a custom CA.
        ignore_https_errors=True,
        client_certificates=[
            {
                "origin": "https://127.0.0.1:8000",
                "certPath": assetdir / "client-certificates/client/trusted/cert.pem",
                "keyPath": assetdir / "client-certificates/client/trusted/key.pem",
            }
        ],
    )
    page = context.new_page()
    page.goto("https://localhost:8000")
    expect(page.get_by_test_id("message")).to_have_text(
        "Sorry, but you need to provide a client certificate to continue."
    )
    page.goto("https://127.0.0.1:8000")
    expect(page.get_by_test_id("message")).to_have_text(
        "Hello Alice, your certificate was issued by localhost!"
    )

    response = page.context.request.get("https://localhost:8000")
    assert (
        "Sorry, but you need to provide a client certificate to continue."
        in response.text()
    )
    response = page.context.request.get("https://127.0.0.1:8000")
    assert "Hello Alice, your certificate was issued by localhost!" in response.text()
    context.close()


def test_should_work_with_new_context_passing_as_content(
    browser: Browser, assetdir: Path
) -> None:
    context = browser.new_context(
        # TODO: Remove this once we can pass a custom CA.
        ignore_https_errors=True,
        client_certificates=[
            {
                "origin": "https://127.0.0.1:8000",
                "cert": (
                    assetdir / "client-certificates/client/trusted/cert.pem"
                ).read_bytes(),
                "key": (
                    assetdir / "client-certificates/client/trusted/key.pem"
                ).read_bytes(),
            }
        ],
    )
    page = context.new_page()
    page.goto("https://localhost:8000")
    expect(page.get_by_test_id("message")).to_have_text(
        "Sorry, but you need to provide a client certificate to continue."
    )
    page.goto("https://127.0.0.1:8000")
    expect(page.get_by_test_id("message")).to_have_text(
        "Hello Alice, your certificate was issued by localhost!"
    )

    response = page.context.request.get("https://localhost:8000")
    assert (
        "Sorry, but you need to provide a client certificate to continue."
        in response.text()
    )
    response = page.context.request.get("https://127.0.0.1:8000")
    assert "Hello Alice, your certificate was issued by localhost!" in response.text()
    context.close()


def test_should_work_with_new_persistent_context(
    browser_type: BrowserType, assetdir: Path, launch_arguments: Dict
) -> None:
    context = browser_type.launch_persistent_context(
        "",
        **launch_arguments,
        # TODO: Remove this once we can pass a custom CA.
        ignore_https_errors=True,
        client_certificates=[
            {
                "origin": "https://127.0.0.1:8000",
                "certPath": assetdir / "client-certificates/client/trusted/cert.pem",
                "keyPath": assetdir / "client-certificates/client/trusted/key.pem",
            }
        ],
    )
    page = context.new_page()
    page.goto("https://localhost:8000")
    expect(page.get_by_test_id("message")).to_have_text(
        "Sorry, but you need to provide a client certificate to continue."
    )
    page.goto("https://127.0.0.1:8000")
    expect(page.get_by_test_id("message")).to_have_text(
        "Hello Alice, your certificate was issued by localhost!"
    )
    context.close()


def test_should_work_with_global_api_request_context(
    playwright: Playwright, assetdir: Path
) -> None:
    request = playwright.request.new_context(
        # TODO: Remove this once we can pass a custom CA.
        ignore_https_errors=True,
        client_certificates=[
            {
                "origin": "https://127.0.0.1:8000",
                "certPath": assetdir / "client-certificates/client/trusted/cert.pem",
                "keyPath": assetdir / "client-certificates/client/trusted/key.pem",
            }
        ],
    )
    response = request.get("https://localhost:8000")
    assert (
        "Sorry, but you need to provide a client certificate to continue."
        in response.text()
    )
    response = request.get("https://127.0.0.1:8000")
    assert "Hello Alice, your certificate was issued by localhost!" in response.text()
    request.dispose()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import List, Optional

from playwright.async_api import Browser, BrowserContext, Request, Route
from tests.server import Server
from tests.utils import must


async def test_link_navigation_inherit_user_agent_from_browser_context(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context(user_agent="hey")

    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    await page.set_content(
        '<a target=_blank rel=noopener href="/popup/popup.html">link</a>'
    )
    request_waitable = asyncio.create_task(server.wait_for_request("/popup/popup.html"))
    await asyncio.sleep(0)  # execute scheduled tasks, but don't await them
    async with context.expect_page() as page_info:
        await page.click("a")
    popup = await page_info.value
    await popup.wait_for_load_state("domcontentloaded")
    user_agent = await popup.evaluate("window.initialUserAgent")
    request = await request_waitable
    assert user_agent == "hey"
    assert request.requestHeaders.getRawHeaders("user-agent") == ["hey"]
    await context.close()


async def test_link_navigation_respect_routes_from_browser_context(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    await page.set_content('<a target=_blank rel=noopener href="empty.html">link</a>')

    intercepted: List[bool] = []

    async def handle_request(route: Route) -> None:
        intercepted.append(True)
        await route.continue_()

    await context.route("**/empty.html", handle_request)
    async with context.expect_page():
        await page.click("a")
    assert intercepted == [True]


async def test_window_open_inherit_user_agent_from_browser_context(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context(user_agent="hey")

    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    request_promise = asyncio.create_task(server.wait_for_request("/dummy.html"))
    await asyncio.sleep(0)  # execute scheduled tasks, but don't await them
    user_agent = await page.evaluate(
        """url => {
            win = window.open(url)
            return win.navigator.userAgent
        }""",
        server.PREFIX + "/dummy.html",
    )
    request = await request_promise
    assert user_agent == "hey"
    assert request.requestHeaders.getRawHeaders("user-agent") == ["hey"]
    await context.close()


async def test_should_inherit_extra_headers_from_browser_context(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context(extra_http_headers={"foo": "bar"})

    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    request_promise = asyncio.create_task(server.wait_for_request("/dummy.html"))
    await asyncio.sleep(0)  # execute scheduled tasks, but don't await them
    await page.evaluate(
        "url => window._popup = window.open(url)", server.PREFIX + "/dummy.html"
    )
    request = await request_promise
    assert request.requestHeaders.getRawHeaders("foo") == ["bar"]
    await context.close()


async def test_should_inherit_offline_from_browser_context(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    await context.set_offline(True)
    online = await page.evaluate(
        """url => {
            win = window.open(url)
            return win.navigator.onLine
        }""",
        server.PREFIX + "/dummy.html",
    )
    assert online is False


async def test_should_inherit_http_credentials_from_browser_context(
    browser: Browser, server: Server
) -> None:
    server.set_auth("/title.html", "user", "pass")
    context = await browser.new_context(
        http_credentials={"username": "user", "password": "pass"}
    )
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_popup() as popup_info:
        await page.evaluate(
            "url => window._popup = window.open(url)", server.PREFIX + "/title.html"
        )
    popup = await popup_info.value
    await popup.wait_for_load_state("domcontentloaded")
    assert await popup.title() == "Woof-Woof"
    await context.close()


async def test_should_inherit_touch_support_from_browser_context(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context(
        viewport={"width": 400, "height": 500}, has_touch=True
    )

    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    has_touch = await page.evaluate(
        """() => {
            win = window.open('')
            return 'ontouchstart' in win
        }"""
    )

    assert has_touch
    await context.close()


async def test_should_inherit_viewport_size_from_browser_context(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context(viewport={"width": 400, "height": 500})

    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    size = await page.evaluate(
        """() => {
            win = window.open('about:blank')
            return { width: win.innerWidth, height: win.innerHeight }
        }"""
    )

    assert size == {"width": 400, "height": 500}
    await context.close()


async def test_should_use_viewport_size_from_window_features(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context(viewport={"width": 700, "height": 700})
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    size = None
    async with page.expect_popup() as popup_info:
        size = await page.evaluate(
            """async () => {
                const win = window.open(window.location.href, 'Title', 'toolbar=no,location=no,directories=no,status=no,menubar=no,scrollbars=yes,resizable=yes,width=600,height=300,top=0,left=0');
                await new Promise(resolve => {
                    const interval = setInterval(() => {
                    if (win.innerWidth === 600 && win.innerHeight === 300) {
                        clearInterval(interval);
                        resolve();
                    }
                    }, 10);
                });
                return { width: win.innerWidth, height: win.innerHeight }
            }"""
        )
    popup = await popup_info.value
    await popup.set_viewport_size({"width": 500, "height": 400})
    await popup.wait_for_load_state()
    resized = await popup.evaluate(
        "() => ({ width: window.innerWidth, height: window.innerHeight })"
    )
    await context.close()
    assert size == {"width": 600, "height": 300}
    assert resized == {"width": 500, "height": 400}


async def test_should_respect_routes_from_browser_context(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)

    def handle_request(route: Route, request: Request, intercepted: List[bool]) -> None:
        asyncio.create_task(route.continue_())
        intercepted.append(True)

    intercepted: List[bool] = []
    await context.route(
        "**/empty.html",
        lambda route, request: handle_request(route, request, intercepted),
    )

    async with page.expect_popup():
        await page.evaluate(
            "url => window.__popup = window.open(url)", server.EMPTY_PAGE
        )
    assert len(intercepted) == 1


async def test_browser_context_add_init_script_should_apply_to_an_in_process_popup(
    context: BrowserContext, server: Server
) -> None:
    await context.add_init_script("window.injected = 123")
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    injected = await page.evaluate(
        """() => {
            const win = window.open('about:blank');
            return win.injected;
        }"""
    )

    assert injected == 123


async def test_browser_context_add_init_script_should_apply_to_a_cross_process_popup(
    context: BrowserContext, server: Server
) -> None:
    await context.add_init_script("window.injected = 123")
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_popup() as popup_info:
        await page.evaluate(
            "url => window.open(url)", server.CROSS_PROCESS_PREFIX + "/title.html"
        )
    popup = await popup_info.value
    assert await popup.evaluate("injected") == 123
    await popup.reload()
    assert await popup.evaluate("injected") == 123


async def test_should_expose_function_from_browser_context(
    context: BrowserContext, server: Server
) -> None:
    await context.expose_function("add", lambda a, b: a + b)
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    added = await page.evaluate(
        """async () => {
            win = window.open('about:blank')
            return win.add(9, 4)
        }"""
    )

    assert added == 13


async def test_should_work(context: BrowserContext) -> None:
    page = await context.new_page()
    async with page.expect_popup() as popup_info:
        await page.evaluate('window.__popup = window.open("about:blank")')
    popup = await popup_info.value
    assert await page.evaluate("!!window.opener") is False
    assert await popup.evaluate("!!window.opener")


async def test_should_work_with_window_features(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_popup() as popup_info:
        await page.evaluate(
            'window.__popup = window.open(window.location.href, "Title", "toolbar=no,location=no,directories=no,status=no,menubar=no,scrollbars=yes,resizable=yes,width=780,height=200,top=0,left=0")'
        )
    popup = await popup_info.value
    assert await page.evaluate("!!window.opener") is False
    assert await popup.evaluate("!!window.opener")


async def test_window_open_emit_for_immediately_closed_popups(
    context: BrowserContext,
) -> None:
    page = await context.new_page()
    async with page.expect_popup() as popup_info:
        await page.evaluate(
            """() => {
                win = window.open('about:blank')
                win.close()
            }"""
        )
    popup = await popup_info.value
    assert popup


async def test_should_emit_for_immediately_closed_popups(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_popup() as popup_info:
        await page.evaluate(
            """() => {
                win = window.open(window.location.href)
                win.close()
            }"""
        )
    popup = await popup_info.value
    assert popup


async def test_should_be_able_to_capture_alert(context: BrowserContext) -> None:
    page = await context.new_page()
    evaluate_task: Optional[asyncio.Future] = None

    async def evaluate() -> None:
        nonlocal evaluate_task
        evaluate_task = asyncio.create_task(
            page.evaluate(
                """() => {
                const win = window.open('')
                win.alert('hello')
            }"""
            )
        )

    [popup, dialog, _] = await asyncio.gather(
        page.wait_for_event("popup"), context.wait_for_event("dialog"), evaluate()
    )

    assert dialog.message == "hello"
    assert dialog.page == popup
    await dialog.dismiss()
    await must(evaluate_task)


async def test_should_work_with_empty_url(context: BrowserContext) -> None:
    page = await context.new_page()
    async with page.expect_popup() as popup_info:
        await page.evaluate("() => window.__popup = window.open('')")
    popup = await popup_info.value
    assert await page.evaluate("!!window.opener") is False
    assert await popup.evaluate("!!window.opener")


async def test_should_work_with_noopener_and_no_url(context: BrowserContext) -> None:
    page = await context.new_page()
    async with page.expect_popup() as popup_info:
        await page.evaluate(
            '() => window.__popup = window.open(undefined, null, "noopener")'
        )
    popup = await popup_info.value
    # Chromium reports 'about:blank#blocked' here.
    assert popup.url.split("#")[0] == "about:blank"
    assert await page.evaluate("!!window.opener") is False
    assert await popup.evaluate("!!window.opener") is False


async def test_should_work_with_noopener_and_about_blank(
    context: BrowserContext,
) -> None:
    page = await context.new_page()
    async with page.expect_popup() as popup_info:
        await page.evaluate(
            '() => window.__popup = window.open("about:blank", null, "noopener")'
        )
    popup = await popup_info.value
    assert await page.evaluate("!!window.opener") is False
    assert await popup.evaluate("!!window.opener") is False


async def test_should_work_with_noopener_and_url(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_popup() as popup_info:
        await page.evaluate(
            'url => window.__popup = window.open(url, null, "noopener")',
            server.EMPTY_PAGE,
        )
    popup = await popup_info.value
    assert await page.evaluate("!!window.opener") is False
    assert await popup.evaluate("!!window.opener") is False


async def test_should_work_with_clicking_target__blank(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    await page.set_content(
        '<a target=_blank rel="opener" href="/one-style.html">yo</a>'
    )
    async with page.expect_popup() as popup_info:
        await page.click("a")
    popup = await popup_info.value
    assert await page.evaluate("!!window.opener") is False
    assert await popup.evaluate("!!window.opener")
    assert popup.main_frame.page == popup


async def test_should_work_with_fake_clicking_target__blank_and_rel_noopener(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    await page.set_content(
        '<a target=_blank rel=noopener href="/one-style.html">yo</a>'
    )
    async with page.expect_popup() as popup_info:
        await page.eval_on_selector("a", "a => a.click()")
    popup = await popup_info.value
    assert await page.evaluate("!!window.opener") is False
    assert await popup.evaluate("!!window.opener") is False


async def test_should_work_with_clicking_target__blank_and_rel_noopener(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    await page.set_content(
        '<a target=_blank rel=noopener href="/one-style.html">yo</a>'
    )
    async with page.expect_popup() as popup_info:
        await page.click("a")
    popup = await popup_info.value
    assert await page.evaluate("!!window.opener") is False
    assert await popup.evaluate("!!window.opener") is False


async def test_should_not_treat_navigations_as_new_popups(
    context: BrowserContext, server: Server
) -> None:
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    await page.set_content(
        '<a target=_blank rel=noopener href="/one-style.html">yo</a>'
    )
    async with page.expect_popup() as popup_info:
        await page.click("a")
    popup = await popup_info.value
    handled_popups = []
    page.on(
        "popup",
        lambda popup: handled_popups.append(True),
    )

    await popup.goto(server.CROSS_PROCESS_PREFIX + "/empty.html")
    assert len(handled_popups) == 0



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from playwright.sync_api import Error, Page


def test_should_wait_for_visible(page: Page) -> None:
    page.set_content('<div id=div style="display:none">content</div>')
    div = page.query_selector("div")
    assert div
    page.evaluate('setTimeout(() => div.style.display = "block", 500)')
    assert div.is_visible() is False
    div.wait_for_element_state("visible")
    assert div.is_visible()


def test_should_wait_for_already_visible(page: Page) -> None:
    page.set_content("<div>content</div>")
    div = page.query_selector("div")
    assert div
    div.wait_for_element_state("visible")


def test_should_timeout_waiting_for_visible(page: Page) -> None:
    page.set_content('<div style="display:none">content</div>')
    div = page.query_selector("div")
    assert div
    with pytest.raises(Error) as exc_info:
        div.wait_for_element_state("visible", timeout=1000)
    assert "Timeout 1000ms exceeded" in exc_info.value.message


def test_should_throw_waiting_for_visible_when_detached(page: Page) -> None:
    page.set_content('<div id=div style="display:none">content</div>')
    div = page.query_selector("div")
    assert div
    page.evaluate("setTimeout(() => div.remove(), 500)")
    with pytest.raises(Error) as exc_info:
        div.wait_for_element_state("visible")
    assert "Element is not attached to the DOM" in exc_info.value.message


def test_should_wait_for_hidden(page: Page) -> None:
    page.set_content("<div id=div>content</div>")
    div = page.query_selector("div")
    assert div
    page.evaluate('setTimeout(() => div.style.display = "none", 500)')
    assert div.is_hidden() is False
    div.wait_for_element_state("hidden")
    assert div.is_hidden()


def test_should_wait_for_already_hidden(page: Page) -> None:
    page.set_content("<div></div>")
    div = page.query_selector("div")
    assert div
    div.wait_for_element_state("hidden")


def test_should_wait_for_hidden_when_detached(page: Page) -> None:
    page.set_content("<div id=div>content</div>")
    div = page.query_selector("div")
    assert div
    page.evaluate("setTimeout(() => div.remove(), 500)")
    div.wait_for_element_state("hidden")
    assert div.is_hidden()


def test_should_wait_for_enabled_button(page: Page) -> None:
    page.set_content("<button id=button disabled><span>Target</span></button>")
    span = page.query_selector("text=Target")
    assert span
    assert span.is_enabled() is False
    page.evaluate("setTimeout(() => button.disabled = false, 500)")
    span.wait_for_element_state("enabled")
    assert span.is_enabled()


def test_should_throw_waiting_for_enabled_when_detached(page: Page) -> None:
    page.set_content("<button id=button disabled>Target</button>")
    button = page.query_selector("button")
    assert button
    page.evaluate("setTimeout(() => button.remove(), 500)")
    with pytest.raises(Error) as exc_info:
        button.wait_for_element_state("enabled")
    assert "Element is not attached to the DOM" in exc_info.value.message


def test_should_wait_for_disabled_button(page: Page) -> None:
    page.set_content("<button id=button><span>Target</span></button>")
    span = page.query_selector("text=Target")
    assert span
    assert span.is_disabled() is False
    page.evaluate("setTimeout(() => button.disabled = true, 500)")
    span.wait_for_element_state("disabled")
    assert span.is_disabled()


def test_should_wait_for_editable_input(page: Page) -> None:
    page.set_content("<input id=input readonly>")
    input = page.query_selector("input")
    assert input
    page.evaluate("setTimeout(() => input.readOnly = false, 500)")
    assert input.is_editable() is False
    input.wait_for_element_state("editable")
    assert input.is_editable()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import datetime
from typing import Callable, List

import pytest

from playwright.async_api import Browser, BrowserContext, Error, Page
from tests.server import Server, TestServerRequest
from tests.utils import must


async def test_should_work(context: BrowserContext, page: Page, server: Server) -> None:
    await page.goto(server.EMPTY_PAGE)
    await context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "password", "value": "123456"}]
    )
    assert await page.evaluate("() => document.cookie") == "password=123456"


async def test_should_roundtrip_cookie(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    # @see https://en.wikipedia.org/wiki/Year_2038_problem
    date = int(datetime.datetime(2038, 1, 1).timestamp() * 1000)
    document_cookie = await page.evaluate(
        """timestamp => {
    const date = new Date(timestamp);
    document.cookie = `username=John Doe;expires=${date.toUTCString()}`;
    return document.cookie;
  }""",
        date,
    )
    assert document_cookie == "username=John Doe"
    cookies = await context.cookies()
    await context.clear_cookies()
    assert await context.cookies() == []
    # TODO: We are waiting for PEP705 so SetCookieParam can be readonly and matches the Cookie type.
    await context.add_cookies(cookies)  # type: ignore
    assert await context.cookies() == cookies


async def test_should_send_cookie_header(
    server: Server, context: BrowserContext
) -> None:
    cookie: List[str] = []

    def handler(request: TestServerRequest) -> None:
        cookie.extend(must(request.requestHeaders.getRawHeaders("cookie")))
        request.finish()

    server.set_route("/empty.html", handler)
    await context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "cookie", "value": "value"}]
    )
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    assert cookie == ["cookie=value"]


async def test_should_isolate_cookies_in_browser_contexts(
    context: BrowserContext, server: Server, browser: Browser
) -> None:
    another_context = await browser.new_context()
    await context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "isolatecookie", "value": "page1value"}]
    )
    await another_context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "isolatecookie", "value": "page2value"}]
    )

    cookies_1 = await context.cookies()
    cookies_2 = await another_context.cookies()
    assert len(cookies_1) == 1
    assert len(cookies_2) == 1
    assert cookies_1[0]["name"] == "isolatecookie"
    assert cookies_1[0]["value"] == "page1value"
    assert cookies_2[0]["name"] == "isolatecookie"
    assert cookies_2[0]["value"] == "page2value"
    await another_context.close()


async def test_should_isolate_session_cookies(
    context: BrowserContext, server: Server, browser: Browser
) -> None:
    server.set_route(
        "/setcookie.html",
        lambda r: (
            r.setHeader("Set-Cookie", "session=value"),
            r.finish(),
        ),
    )

    page_1 = await context.new_page()
    await page_1.goto(server.PREFIX + "/setcookie.html")
    ##
    page_2 = await context.new_page()
    await page_2.goto(server.EMPTY_PAGE)
    cookies_2 = await context.cookies()
    assert len(cookies_2) == 1
    assert ",".join(list(map(lambda c: c["value"], cookies_2))) == "value"
    ##
    context_b = await browser.new_context()
    page_3 = await context_b.new_page()
    await page_3.goto(server.EMPTY_PAGE)
    cookies_3 = await context_b.cookies()
    assert cookies_3 == []
    await context_b.close()


async def test_should_isolate_persistent_cookies(
    context: BrowserContext, server: Server, browser: Browser
) -> None:
    server.set_route(
        "/setcookie.html",
        lambda r: (
            r.setHeader("Set-Cookie", "persistent=persistent-value; max-age=3600"),
            r.finish(),
        ),
    )

    page = await context.new_page()
    await page.goto(server.PREFIX + "/setcookie.html")

    context_1 = context
    context_2 = await browser.new_context()
    [page_1, page_2] = await asyncio.gather(context_1.new_page(), context_2.new_page())
    await asyncio.gather(page_1.goto(server.EMPTY_PAGE), page_2.goto(server.EMPTY_PAGE))
    [cookies_1, cookies_2] = await asyncio.gather(
        context_1.cookies(), context_2.cookies()
    )
    assert len(cookies_1) == 1
    assert cookies_1[0]["name"] == "persistent"
    assert cookies_1[0]["value"] == "persistent-value"
    assert len(cookies_2) == 0
    await context_2.close()


async def test_should_isolate_send_cookie_header(
    server: Server, context: BrowserContext, browser: Browser
) -> None:
    cookie: List[str] = []

    def handler(request: TestServerRequest) -> None:
        cookie.extend(request.requestHeaders.getRawHeaders("cookie") or [])
        request.finish()

    server.set_route("/empty.html", handler)

    await context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "sendcookie", "value": "value"}]
    )

    page_1 = await context.new_page()
    await page_1.goto(server.EMPTY_PAGE)
    assert cookie == ["sendcookie=value"]
    cookie.clear()
    ##
    context_2 = await browser.new_context()
    page_2 = await context_2.new_page()
    await page_2.goto(server.EMPTY_PAGE)
    assert cookie == []
    await context_2.close()


async def test_should_isolate_cookies_between_launches(
    browser_factory: Callable[..., "asyncio.Future[Browser]"], server: Server
) -> None:
    browser_1 = await browser_factory()
    context_1 = await browser_1.new_context()
    await context_1.add_cookies(
        [
            {
                "url": server.EMPTY_PAGE,
                "name": "cookie-in-context-1",
                "value": "value",
                "expires": int(datetime.datetime.now().timestamp() + 10000),
            }
        ]
    )
    await browser_1.close()

    browser_2 = await browser_factory()
    context_2 = await browser_2.new_context()
    cookies = await context_2.cookies()
    assert cookies == []
    await browser_2.close()


async def test_should_set_multiple_cookies(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    await context.add_cookies(
        [
            {"url": server.EMPTY_PAGE, "name": "multiple-1", "value": "123456"},
            {"url": server.EMPTY_PAGE, "name": "multiple-2", "value": "bar"},
        ]
    )
    assert (
        await page.evaluate(
            """() => {
    const cookies = document.cookie.split(';');
    return cookies.map(cookie => cookie.trim()).sort();
  }"""
        )
        == ["multiple-1=123456", "multiple-2=bar"]
    )


async def test_should_have_expires_set_to_neg_1_for_session_cookies(
    context: BrowserContext, server: Server
) -> None:
    await context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "expires", "value": "123456"}]
    )
    cookies = await context.cookies()
    assert cookies[0]["expires"] == -1


async def test_should_set_cookie_with_reasonable_defaults(
    context: BrowserContext,
    server: Server,
    default_same_site_cookie_value: str,
) -> None:
    await context.add_cookies(
        [{"url": server.EMPTY_PAGE, "name": "defaults", "value": "123456"}]
    )
    cookies = await context.cookies()
    cookies.sort(key=lambda r: r["name"])
    assert cookies == [
        {
            "name": "defaults",
            "value": "123456",
            "domain": "localhost",
            "path": "/",
            "expires": -1,
            "httpOnly": False,
            "secure": False,
            "sameSite": default_same_site_cookie_value,
        }
    ]


async def test_should_set_a_cookie_with_a_path(
    context: BrowserContext,
    page: Page,
    server: Server,
    default_same_site_cookie_value: str,
) -> None:
    await page.goto(server.PREFIX + "/grid.html")
    await context.add_cookies(
        [
            {
                "domain": "localhost",
                "path": "/grid.html",
                "name": "gridcookie",
                "value": "GRID",
            }
        ]
    )
    assert await context.cookies() == [
        {
            "name": "gridcookie",
            "value": "GRID",
            "domain": "localhost",
            "path": "/grid.html",
            "expires": -1,
            "httpOnly": False,
            "secure": False,
            "sameSite": default_same_site_cookie_value,
        }
    ]
    assert await page.evaluate("document.cookie") == "gridcookie=GRID"
    await page.goto(server.EMPTY_PAGE)
    assert await page.evaluate("document.cookie") == ""
    await page.goto(server.PREFIX + "/grid.html")
    assert await page.evaluate("document.cookie") == "gridcookie=GRID"


async def test_should_not_set_a_cookie_with_blank_page_url(
    context: BrowserContext, server: Server
) -> None:
    with pytest.raises(Error) as exc_info:
        await context.add_cookies(
            [
                {"url": server.EMPTY_PAGE, "name": "example-cookie", "value": "best"},
                {"url": "about:blank", "name": "example-cookie-blank", "value": "best"},
            ]
        )
    assert (
        'Blank page can not have cookie "example-cookie-blank"'
        in exc_info.value.message
    )


async def test_should_not_set_a_cookie_on_a_data_url_page(
    context: BrowserContext,
) -> None:
    with pytest.raises(Error) as exc_info:
        await context.add_cookies(
            [
                {
                    "url": "data:,Hello%2C%20World!",
                    "name": "example-cookie",
                    "value": "best",
                }
            ]
        )
    assert (
        'Data URL page can not have cookie "example-cookie"' in exc_info.value.message
    )


async def test_should_default_to_setting_secure_cookie_for_https_websites(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    SECURE_URL = "https://example.com"
    await context.add_cookies([{"url": SECURE_URL, "name": "foo", "value": "bar"}])
    [cookie] = await context.cookies(SECURE_URL)
    assert cookie["secure"]


async def test_should_be_able_to_set_unsecure_cookie_for_http_website(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    HTTP_URL = "http://example.com"
    await context.add_cookies([{"url": HTTP_URL, "name": "foo", "value": "bar"}])
    [cookie] = await context.cookies(HTTP_URL)
    assert not cookie["secure"]


async def test_should_set_a_cookie_on_a_different_domain(
    context: BrowserContext,
    page: Page,
    server: Server,
    default_same_site_cookie_value: str,
) -> None:
    await page.goto(server.EMPTY_PAGE)
    await context.add_cookies(
        [{"url": "https://www.example.com", "name": "example-cookie", "value": "best"}]
    )
    assert await page.evaluate("document.cookie") == ""
    assert await context.cookies("https://www.example.com") == [
        {
            "name": "example-cookie",
            "value": "best",
            "domain": "www.example.com",
            "path": "/",
            "expires": -1,
            "httpOnly": False,
            "secure": True,
            "sameSite": default_same_site_cookie_value,
        }
    ]


async def test_should_set_cookies_for_a_frame(
    context: BrowserContext, page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    await context.add_cookies(
        [{"url": server.PREFIX, "name": "frame-cookie", "value": "value"}]
    )
    await page.evaluate(
        """src => {
    let fulfill;
    const promise = new Promise(x => fulfill = x);
    const iframe = document.createElement('iframe');
    document.body.appendChild(iframe);
    iframe.onload = fulfill;
    iframe.src = src;
    return promise;
  }""",
        server.PREFIX + "/grid.html",
    )

    assert await page.frames[1].evaluate("document.cookie") == "frame-cookie=value"


async def test_should_not_block_third_party_cookies(
    context: BrowserContext,
    page: Page,
    server: Server,
    is_chromium: bool,
    is_firefox: bool,
) -> None:
    await page.goto(server.EMPTY_PAGE)
    await page.evaluate(
        """src => {
    let fulfill;
    const promise = new Promise(x => fulfill = x);
    const iframe = document.createElement('iframe');
    document.body.appendChild(iframe);
    iframe.onload = fulfill;
    iframe.src = src;
    return promise;
  }""",
        server.CROSS_PROCESS_PREFIX + "/grid.html",
    )
    await page.frames[1].evaluate("document.cookie = 'username=John Doe'")
    await page.wait_for_timeout(2000)
    allows_third_party = is_firefox
    cookies = await context.cookies(server.CROSS_PROCESS_PREFIX + "/grid.html")

    if allows_third_party:
        assert cookies == [
            {
                "domain": "127.0.0.1",
                "expires": -1,
                "httpOnly": False,
                "name": "username",
                "path": "/",
                "sameSite": "Lax" if is_chromium else "None",
                "secure": False,
                "value": "John Doe",
            }
        ]
    else:
        assert cookies == []



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from playwright.sync_api import Page, Response
from tests.server import Server


def test_listeners(page: Page, server: Server) -> None:
    log = []

    def print_response(response: Response) -> None:
        log.append(response)

    page.on("response", print_response)
    page.goto(f"{server.PREFIX}/input/textarea.html")
    assert len(log) > 0
    page.remove_listener("response", print_response)

    log = []
    page.goto(f"{server.PREFIX}/input/textarea.html")
    assert len(log) == 0



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Generator

import pytest

from playwright.sync_api import Page, expect

from .utils import TODO_ITEMS, assert_number_of_todos_in_local_storage


@pytest.fixture(autouse=True)
def run_around_tests(page: Page) -> Generator[None, None, None]:
    # setup before a test
    page.goto("https://demo.playwright.dev/todomvc")
    # run the actual test
    yield
    # run any cleanup code


def test_should_display_the_current_number_of_todo_items(page: Page) -> None:
    page.locator(".new-todo").fill(TODO_ITEMS[0])
    page.locator(".new-todo").press("Enter")
    expect(page.locator(".todo-count")).to_contain_text("1")

    page.locator(".new-todo").fill(TODO_ITEMS[1])
    page.locator(".new-todo").press("Enter")
    expect(page.locator(".todo-count")).to_contain_text("2")

    assert_number_of_todos_in_local_storage(page, 2)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Optional

from playwright.async_api import Page, Route
from tests.server import Server, TestServerRequest


async def test_request_continue_should_work(page: Page, server: Server) -> None:
    await page.route("**/*", lambda route: asyncio.create_task(route.continue_()))
    await page.goto(server.EMPTY_PAGE)


async def test_request_continue_should_amend_http_headers(
    page: Page, server: Server
) -> None:
    await page.route(
        "**/*",
        lambda route: asyncio.create_task(
            route.continue_(headers={**route.request.headers, "FOO": "bar"})
        ),
    )

    await page.goto(server.EMPTY_PAGE)
    [request, _] = await asyncio.gather(
        server.wait_for_request("/sleep.zzz"),
        page.evaluate('() => fetch("/sleep.zzz")'),
    )
    assert request.getHeader("foo") == "bar"


async def test_request_continue_should_amend_method(page: Page, server: Server) -> None:
    server_request = asyncio.create_task(server.wait_for_request("/sleep.zzz"))
    await page.goto(server.EMPTY_PAGE)
    await page.route(
        "**/*", lambda route: asyncio.create_task(route.continue_(method="POST"))
    )
    [request, _] = await asyncio.gather(
        server.wait_for_request("/sleep.zzz"),
        page.evaluate('() => fetch("/sleep.zzz")'),
    )
    assert request.method.decode() == "POST"
    assert (await server_request).method.decode() == "POST"


async def test_request_continue_should_amend_method_on_main_request(
    page: Page, server: Server
) -> None:
    request = asyncio.create_task(server.wait_for_request("/empty.html"))
    await page.route(
        "**/*", lambda route: asyncio.create_task(route.continue_(method="POST"))
    )
    await page.goto(server.EMPTY_PAGE)
    assert (await request).method.decode() == "POST"


async def test_request_continue_should_amend_post_data(
    page: Page, server: Server
) -> None:
    await page.goto(server.EMPTY_PAGE)
    await page.route(
        "**/*",
        lambda route: asyncio.create_task(route.continue_(post_data=b"doggo")),
    )

    [server_request, _] = await asyncio.gather(
        server.wait_for_request("/sleep.zzz"),
        page.evaluate(
            """
            () => fetch('/sleep.zzz', { method: 'POST', body: 'birdy' })
            """
        ),
    )
    assert server_request.post_body
    assert server_request.post_body.decode() == "doggo"


async def test_should_override_request_url(page: Page, server: Server) -> None:
    request = asyncio.create_task(server.wait_for_request("/empty.html"))
    await page.route(
        "**/foo",
        lambda route: asyncio.create_task(route.continue_(url=server.EMPTY_PAGE)),
    )

    await page.goto(server.PREFIX + "/foo")
    assert (await request).method == b"GET"


async def test_should_raise_except(page: Page, server: Server) -> None:
    exc_fut: "asyncio.Future[Optional[Exception]]" = asyncio.Future()

    async def capture_exception(route: Route) -> None:
        try:
            await route.continue_(url="file:///tmp/does-not-exist")
            exc_fut.set_result(None)
        except Exception as e:
            exc_fut.set_result(e)

    await page.route("**/*", capture_exception)
    asyncio.create_task(page.goto(server.EMPTY_PAGE))
    assert "New URL must have same protocol as overridden URL" in str(await exc_fut)


async def test_should_amend_utf8_post_data(page: Page, server: Server) -> None:
    await page.goto(server.EMPTY_PAGE)
    await page.route(
        "**/*",
        lambda route: asyncio.create_task(route.continue_(post_data="пушкин")),
    )

    [server_request, result] = await asyncio.gather(
        server.wait_for_request("/sleep.zzz"),
        page.evaluate("fetch('/sleep.zzz', { method: 'POST', body: 'birdy' })"),
    )
    assert server_request.method == b"POST"
    assert server_request.post_body
    assert server_request.post_body.decode("utf8") == "пушкин"


async def test_should_amend_binary_post_data(page: Page, server: Server) -> None:
    await page.goto(server.EMPTY_PAGE)
    await page.route(
        "**/*",
        lambda route: asyncio.create_task(
            route.continue_(post_data=b"\x00\x01\x02\x03\x04")
        ),
    )

    [server_request, result] = await asyncio.gather(
        server.wait_for_request("/sleep.zzz"),
        page.evaluate("fetch('/sleep.zzz', { method: 'POST', body: 'birdy' })"),
    )
    assert server_request.method == b"POST"
    assert server_request.post_body == b"\x00\x01\x02\x03\x04"


async def test_continue_should_not_change_multipart_form_data_body(
    page: Page, server: Server, browser_name: str
) -> None:
    await page.goto(server.EMPTY_PAGE)
    server.set_route(
        "/upload",
        lambda context: (
            context.write(b"done"),
            context.setHeader("Content-Type", "text/plain"),
            context.finish(),
        ),
    )

    async def send_form_data() -> TestServerRequest:
        req_task = asyncio.create_task(server.wait_for_request("/upload"))
        status = await page.evaluate(
            """async () => {
            const newFile = new File(['file content'], 'file.txt');
            const formData = new FormData();
            formData.append('file', newFile);
            const response = await fetch('/upload', {
                method: 'POST',
                credentials: 'include',
                body: formData,
            });
            return response.status;
        }"""
        )
        req = await req_task
        assert status == 200
        return req

    req_before = await send_form_data()
    await page.route("**/*", lambda route: route.continue_())
    req_after = await send_form_data()

    file_content = (
        'Content-Disposition: form-data; name="file"; filename="file.txt"\r\n'
        "Content-Type: application/octet-stream\r\n"
        "\r\n"
        "file content\r\n"
        "------"
    )
    assert req_before.post_body
    assert req_after.post_body
    assert file_content in req_before.post_body.decode()
    assert file_content in req_after.post_body.decode()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from playwright.async_api import Browser, BrowserContext, Error, Page
from tests.server import Server


async def test_should_work(page: Page, server: Server, context: BrowserContext) -> None:
    await context.grant_permissions(["geolocation"])
    await page.goto(server.EMPTY_PAGE)
    await context.set_geolocation({"latitude": 10, "longitude": 10})
    geolocation = await page.evaluate(
        """() => new Promise(resolve => navigator.geolocation.getCurrentPosition(position => {
      resolve({latitude: position.coords.latitude, longitude: position.coords.longitude});
    }))"""
    )
    assert geolocation == {"latitude": 10, "longitude": 10}


async def test_should_throw_when_invalid_longitude(context: BrowserContext) -> None:
    with pytest.raises(Error) as exc:
        await context.set_geolocation({"latitude": 10, "longitude": 200})
    assert (
        "geolocation.longitude: precondition -180 <= LONGITUDE <= 180 failed."
        in exc.value.message
    )


async def test_should_isolate_contexts(
    page: Page, server: Server, context: BrowserContext, browser: Browser
) -> None:
    await context.grant_permissions(["geolocation"])
    await context.set_geolocation({"latitude": 10, "longitude": 10})
    await page.goto(server.EMPTY_PAGE)

    context2 = await browser.new_context(
        permissions=["geolocation"], geolocation={"latitude": 10.5, "longitude": 10.5}
    )

    page2 = await context2.new_page()
    await page2.goto(server.EMPTY_PAGE)

    geolocation = await page.evaluate(
        """() => new Promise(resolve => navigator.geolocation.getCurrentPosition(position => {
      resolve({latitude: position.coords.latitude, longitude: position.coords.longitude})
    }))"""
    )
    assert geolocation == {"latitude": 10, "longitude": 10}

    geolocation2 = await page2.evaluate(
        """() => new Promise(resolve => navigator.geolocation.getCurrentPosition(position => {
      resolve({latitude: position.coords.latitude, longitude: position.coords.longitude})
    }))"""
    )
    assert geolocation2 == {"latitude": 10.5, "longitude": 10.5}

    await context2.close()


async def test_should_use_context_options(browser: Browser, server: Server) -> None:
    context = await browser.new_context(
        geolocation={"latitude": 10, "longitude": 10}, permissions=["geolocation"]
    )
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)

    geolocation = await page.evaluate(
        """() => new Promise(resolve => navigator.geolocation.getCurrentPosition(position => {
      resolve({latitude: position.coords.latitude, longitude: position.coords.longitude});
    }))"""
    )
    assert geolocation == {"latitude": 10, "longitude": 10}
    await context.close()


async def test_watch_position_should_be_notified(
    page: Page, server: Server, context: BrowserContext
) -> None:
    await context.grant_permissions(["geolocation"])
    await page.goto(server.EMPTY_PAGE)
    messages = []
    page.on("console", lambda message: messages.append(message.text))

    await context.set_geolocation({"latitude": 0, "longitude": 0})
    await page.evaluate(
        """() => {
      navigator.geolocation.watchPosition(pos => {
        const coords = pos.coords;
        console.log(`lat=${coords.latitude} lng=${coords.longitude}`);
      }, err => {});
    }"""
    )

    async with page.expect_console_message(lambda m: "lat=0 lng=10" in m.text):
        await context.set_geolocation({"latitude": 0, "longitude": 10})

    async with page.expect_console_message(lambda m: "lat=20 lng=30" in m.text):
        await context.set_geolocation({"latitude": 20, "longitude": 30})

    async with page.expect_console_message(lambda m: "lat=40 lng=50" in m.text):
        await context.set_geolocation({"latitude": 40, "longitude": 50})

    all_messages = "|".join(messages)
    assert "lat=0 lng=10" in all_messages
    assert "lat=20 lng=30" in all_messages
    assert "lat=40 lng=50" in all_messages


async def test_should_use_context_options_for_popup(
    page: Page, context: BrowserContext, server: Server
) -> None:
    await context.grant_permissions(["geolocation"])
    await context.set_geolocation({"latitude": 10, "longitude": 10})
    async with page.expect_popup() as popup_info:
        await page.evaluate(
            "url => window._popup = window.open(url)",
            server.PREFIX + "/geolocation.html",
        )
    popup = await popup_info.value
    await popup.wait_for_load_state()
    geolocation = await popup.evaluate("() => window.geolocationPromise")
    assert geolocation == {"latitude": 10, "longitude": 10}



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from playwright.async_api import Browser, Error
from tests.server import Server


async def test_ignore_https_error_should_work(
    browser: Browser, https_server: Server
) -> None:
    context = await browser.new_context(ignore_https_errors=True)
    page = await context.new_page()
    response = await page.goto(https_server.EMPTY_PAGE)
    assert response
    assert response.ok
    await context.close()


async def test_ignore_https_error_should_work_negative_case(
    browser: Browser, https_server: Server
) -> None:
    context = await browser.new_context()
    page = await context.new_page()
    with pytest.raises(Error):
        await page.goto(https_server.EMPTY_PAGE)
    await context.close()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import base64
import json
from typing import Any, Callable, cast
from urllib.parse import parse_qs

import pytest

from playwright.async_api import Browser, BrowserContext, Error, FilePayload, Page
from tests.server import Server, TestServerRequest
from tests.utils import must


async def test_get_should_work(context: BrowserContext, server: Server) -> None:
    response = await context.request.get(server.PREFIX + "/simple.json")
    assert response.url == server.PREFIX + "/simple.json"
    assert response.status == 200
    assert response.status_text == "OK"
    assert response.ok is True
    assert response.headers["content-type"] == "application/json"
    assert {
        "name": "Content-Type",
        "value": "application/json",
    } in response.headers_array
    assert await response.text() == '{"foo": "bar"}\n'


async def test_fetch_should_work(context: BrowserContext, server: Server) -> None:
    response = await context.request.fetch(server.PREFIX + "/simple.json")
    assert response.url == server.PREFIX + "/simple.json"
    assert response.status == 200
    assert response.status_text == "OK"
    assert response.ok is True
    assert response.headers["content-type"] == "application/json"
    assert {
        "name": "Content-Type",
        "value": "application/json",
    } in response.headers_array
    assert await response.text() == '{"foo": "bar"}\n'


async def test_should_throw_on_network_error(
    context: BrowserContext, server: Server
) -> None:
    server.set_route("/test", lambda request: request.loseConnection())
    with pytest.raises(Error, match="socket hang up"):
        await context.request.fetch(server.PREFIX + "/test")


async def test_should_add_session_cookies_to_request(
    context: BrowserContext, server: Server
) -> None:
    await context.add_cookies(
        [
            {
                "name": "username",
                "value": "John Doe",
                "url": server.EMPTY_PAGE,
                "expires": -1,
                "httpOnly": False,
                "secure": False,
                "sameSite": "Lax",
            }
        ]
    )
    [server_req, response] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        context.request.get(server.EMPTY_PAGE),
    )
    assert server_req.getHeader("Cookie") == "username=John Doe"


@pytest.mark.parametrize(
    "method", ["fetch", "delete", "get", "head", "patch", "post", "put"]
)
async def test_should_support_query_params(
    context: BrowserContext, server: Server, method: str
) -> None:
    expected_params = {"p1": "v1", "парам2": "знач2"}
    [server_req, _] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        getattr(context.request, method)(
            server.EMPTY_PAGE + "?p1=foo", params=expected_params
        ),
    )
    assert list(map(lambda x: x.decode(), server_req.args["p1".encode()])) == [
        "foo",
        "v1",
    ]
    assert server_req.args["парам2".encode()][0].decode() == "знач2"


@pytest.mark.parametrize(
    "method", ["fetch", "delete", "get", "head", "patch", "post", "put"]
)
async def test_should_support_params_passed_as_object(
    context: BrowserContext, server: Server, method: str
) -> None:
    params = {
        "param1": "value1",
        "парам2": "знач2",
    }
    [server_req, _] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        getattr(context.request, method)(server.EMPTY_PAGE, params=params),
    )
    assert server_req.args["param1".encode()][0].decode() == "value1"
    assert len(server_req.args["param1".encode()]) == 1
    assert server_req.args["парам2".encode()][0].decode() == "знач2"


@pytest.mark.parametrize(
    "method", ["fetch", "delete", "get", "head", "patch", "post", "put"]
)
async def test_should_support_params_passed_as_strings(
    context: BrowserContext, server: Server, method: str
) -> None:
    params = "?param1=value1&param1=value2&парам2=знач2"
    [server_req, _] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        getattr(context.request, method)(server.EMPTY_PAGE, params=params),
    )
    assert list(map(lambda x: x.decode(), server_req.args["param1".encode()])) == [
        "value1",
        "value2",
    ]
    assert len(server_req.args["param1".encode()]) == 2
    assert server_req.args["парам2".encode()][0].decode() == "знач2"


@pytest.mark.parametrize(
    "method", ["fetch", "delete", "get", "head", "patch", "post", "put"]
)
async def test_should_support_fail_on_status_code(
    context: BrowserContext, server: Server, method: str
) -> None:
    with pytest.raises(Error, match="404 Not Found"):
        await getattr(context.request, method)(
            server.PREFIX + "/this-does-clearly-not-exist.html",
            fail_on_status_code=True,
        )


@pytest.mark.parametrize(
    "method", ["fetch", "delete", "get", "head", "patch", "post", "put"]
)
async def test_should_support_ignore_https_errors_option(
    context: BrowserContext, https_server: Server, method: str
) -> None:
    response = await getattr(context.request, method)(
        https_server.EMPTY_PAGE, ignore_https_errors=True
    )
    assert response.ok
    assert response.status == 200


async def test_should_not_add_context_cookie_if_cookie_header_passed_as_parameter(
    context: BrowserContext, server: Server
) -> None:
    await context.add_cookies(
        [
            {
                "name": "username",
                "value": "John Doe",
                "url": server.EMPTY_PAGE,
                "expires": -1,
                "httpOnly": False,
                "secure": False,
                "sameSite": "Lax",
            }
        ]
    )
    [server_req, _] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        context.request.get(server.EMPTY_PAGE, headers={"Cookie": "foo=bar"}),
    )
    assert server_req.getHeader("Cookie") == "foo=bar"


async def test_should_support_http_credentials_send_immediately_for_browser_context(
    context_factory: "Callable[..., asyncio.Future[BrowserContext]]", server: Server
) -> None:
    context = await context_factory(
        http_credentials={
            "username": "user",
            "password": "pass",
            "origin": server.PREFIX.upper(),
            "send": "always",
        }
    )
    # First request
    server_request, response = await asyncio.gather(
        server.wait_for_request("/empty.html"), context.request.get(server.EMPTY_PAGE)
    )
    expected_auth = "Basic " + base64.b64encode(b"user:pass").decode()
    assert server_request.getHeader("authorization") == expected_auth
    assert response.status == 200

    # Second request
    server_request, response = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        context.request.get(server.CROSS_PROCESS_PREFIX + "/empty.html"),
    )
    # Not sent to another origin.
    assert server_request.getHeader("authorization") is None
    assert response.status == 200


async def test_support_http_credentials_send_immediately_for_browser_new_page(
    server: Server, browser: Browser
) -> None:
    page = await browser.new_page(
        http_credentials={
            "username": "user",
            "password": "pass",
            "origin": server.PREFIX.upper(),
            "send": "always",
        }
    )
    server_request, response = await asyncio.gather(
        server.wait_for_request("/empty.html"), page.request.get(server.EMPTY_PAGE)
    )
    assert (
        server_request.getHeader("authorization")
        == "Basic " + base64.b64encode(b"user:pass").decode()
    )
    assert response.status == 200

    server_request, response = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        page.request.get(server.CROSS_PROCESS_PREFIX + "/empty.html"),
    )
    # Not sent to another origin.
    assert server_request.getHeader("authorization") is None
    assert response.status == 200

    await page.close()


@pytest.mark.parametrize("method", ["delete", "patch", "post", "put"])
async def test_should_support_post_data(
    context: BrowserContext, method: str, server: Server
) -> None:
    async def support_post_data(fetch_data: Any, request_post_data: Any) -> None:
        [request, response] = await asyncio.gather(
            server.wait_for_request("/simple.json"),
            getattr(context.request, method)(
                server.PREFIX + "/simple.json", data=fetch_data
            ),
        )
        assert request.method.decode() == method.upper()
        assert request.post_body == request_post_data
        assert response.status == 200
        assert response.url == server.PREFIX + "/simple.json"
        assert request.getHeader("Content-Length") == str(len(must(request.post_body)))

    await support_post_data("My request", "My request".encode())
    await support_post_data(b"My request", "My request".encode())
    await support_post_data(["my", "request"], json.dumps(["my", "request"]).encode())
    await support_post_data({"my": "request"}, json.dumps({"my": "request"}).encode())
    with pytest.raises(Error, match="Unsupported 'data' type: <class 'function'>"):
        await support_post_data(lambda: None, None)


async def test_should_support_application_x_www_form_urlencoded(
    context: BrowserContext, server: Server
) -> None:
    [request, response] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        context.request.post(
            server.PREFIX + "/empty.html",
            form={
                "firstName": "John",
                "lastName": "Doe",
                "file": "f.js",
            },
        ),
    )
    assert request.method == b"POST"
    assert request.getHeader("Content-Type") == "application/x-www-form-urlencoded"
    assert request.post_body
    body = request.post_body.decode()
    assert request.getHeader("Content-Length") == str(len(body))
    params = parse_qs(request.post_body)
    assert params[b"firstName"] == [b"John"]
    assert params[b"lastName"] == [b"Doe"]
    assert params[b"file"] == [b"f.js"]


async def test_should_support_multipart_form_data(
    context: BrowserContext, server: Server
) -> None:
    file: FilePayload = {
        "name": "f.js",
        "mimeType": "text/javascript",
        "buffer": b"var x = 10;\r\n;console.log(x);",
    }
    [request, _] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        context.request.post(
            server.PREFIX + "/empty.html",
            multipart={
                "firstName": "John",
                "lastName": "Doe",
                "file": file,
            },
        ),
    )
    assert request.method == b"POST"
    assert cast(str, request.getHeader("Content-Type")).startswith(
        "multipart/form-data; "
    )
    assert must(request.getHeader("Content-Length")) == str(
        len(must(request.post_body))
    )
    assert request.args[b"firstName"] == [b"John"]
    assert request.args[b"lastName"] == [b"Doe"]
    assert request.args[b"file"][0] == file["buffer"]


async def test_should_add_default_headers(
    context: BrowserContext, page: Page, server: Server
) -> None:
    [request, response] = await asyncio.gather(
        server.wait_for_request("/empty.html"),
        context.request.get(server.EMPTY_PAGE),
    )
    assert request.getHeader("Accept") == "*/*"
    assert request.getHeader("Accept-Encoding") == "gzip,deflate,br"
    assert request.getHeader("User-Agent") == await page.evaluate(
        "() => navigator.userAgent"
    )


async def test_should_work_after_context_dispose(
    context: BrowserContext, server: Server
) -> None:
    await context.close(reason="Test ended.")
    with pytest.raises(Error, match="Test ended."):
        await context.request.get(server.EMPTY_PAGE)


async def test_should_retry_ECONNRESET(context: BrowserContext, server: Server) -> None:
    request_count = 0

    def _handle_request(req: TestServerRequest) -> None:
        nonlocal request_count
        request_count += 1
        if request_count <= 3:
            assert req.transport
            req.transport.abortConnection()
            return
        req.setHeader("content-type", "text/plain")
        req.write(b"Hello!")
        req.finish()

    server.set_route("/test", _handle_request)
    response = await context.request.fetch(server.PREFIX + "/test", max_retries=3)
    assert response.status == 200
    assert await response.text() == "Hello!"
    assert request_count == 4



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from typing import Optional

import pytest

from playwright.async_api import BrowserContext, Page
from tests.utils import must

from ..server import Server, TestServerRequest


async def test_console_event_should_work(page: Page) -> None:
    [message, _] = await asyncio.gather(
        page.context.wait_for_event("console"),
        page.evaluate("() => console.log('hello')"),
    )
    assert message.text == "hello"
    assert message.page == page


async def test_console_event_should_work_in_popup(page: Page) -> None:
    [message, popup, _] = await asyncio.gather(
        page.context.wait_for_event("console"),
        page.wait_for_event("popup"),
        page.evaluate(
            """() => {
            const win = window.open('');
            win.console.log('hello');
        }"""
        ),
    )
    assert message.text == "hello"
    assert message.page == popup


# console message from javascript: url is not reported at all
@pytest.mark.skip_browser("firefox")
async def test_console_event_should_work_in_popup_2(
    page: Page, browser_name: str
) -> None:
    [message, popup, _] = await asyncio.gather(
        page.context.wait_for_event("console", lambda msg: msg.type == "log"),
        page.context.wait_for_event("page"),
        page.evaluate(
            """async () => {
            const win = window.open('javascript:console.log("hello")');
            await new Promise(f => setTimeout(f, 0));
            win.close();
        }"""
        ),
    )
    assert message.text == "hello"
    assert message.page == popup


# console message from javascript: url is not reported at all
@pytest.mark.skip_browser("firefox")
async def test_console_event_should_work_in_immediately_closed_popup(
    page: Page, browser_name: str
) -> None:
    [message, popup, _] = await asyncio.gather(
        page.context.wait_for_event("console"),
        page.wait_for_event("popup"),
        page.evaluate(
            """async () => {
            const win = window.open();
            win.console.log('hello');
            win.close();
        }"""
        ),
    )
    assert message.text == "hello"
    assert message.page == popup


async def test_dialog_event_should_work1(page: Page) -> None:
    prompt_task: Optional[asyncio.Future[str]] = None

    async def open_dialog() -> None:
        nonlocal prompt_task
        prompt_task = asyncio.create_task(page.evaluate("() => prompt('hey?')"))

    [dialog1, dialog2, _] = await asyncio.gather(
        page.context.wait_for_event("dialog"),
        page.wait_for_event("dialog"),
        open_dialog(),
    )
    assert dialog1 == dialog2
    assert dialog1.message == "hey?"
    assert dialog1.page == page
    await dialog1.accept("hello")
    assert await must(prompt_task) == "hello"


async def test_dialog_event_should_work_in_popup(page: Page) -> None:
    prompt_task: Optional[asyncio.Future[str]] = None

    async def open_dialog() -> None:
        nonlocal prompt_task
        prompt_task = asyncio.create_task(
            page.evaluate("() => window.open('').prompt('hey?')")
        )

    [dialog, popup, _] = await asyncio.gather(
        page.context.wait_for_event("dialog"),
        page.wait_for_event("popup"),
        open_dialog(),
    )
    assert dialog.message == "hey?"
    assert dialog.page == popup
    await dialog.accept("hello")
    assert await must(prompt_task) == "hello"


# console message from javascript: url is not reported at all
@pytest.mark.skip_browser("firefox")
async def test_dialog_event_should_work_in_popup_2(
    page: Page, browser_name: str
) -> None:
    promise = asyncio.create_task(
        page.evaluate("() => window.open('javascript:prompt(\"hey?\")')")
    )
    dialog = await page.context.wait_for_event("dialog")
    assert dialog.message == "hey?"
    assert dialog.page is None
    await dialog.accept("hello")
    await promise


# console message from javascript: url is not reported at all
@pytest.mark.skip_browser("firefox")
async def test_dialog_event_should_work_in_immdiately_closed_popup(page: Page) -> None:
    [message, popup, _] = await asyncio.gather(
        page.context.wait_for_event("console"),
        page.wait_for_event("popup"),
        page.evaluate(
            """() => {
            const win = window.open();
            win.console.log('hello');
            win.close();
        }"""
        ),
    )
    assert message.text == "hello"
    assert message.page == popup


async def test_dialog_event_should_work_with_inline_script_tag(
    page: Page, server: Server
) -> None:
    def handle_route(request: TestServerRequest) -> None:
        request.setHeader("content-type", "text/html")
        request.write(b"""<script>window.result = prompt('hey?')</script>""")
        request.finish()

    server.set_route("/popup.html", handle_route)
    await page.goto(server.EMPTY_PAGE)
    await page.set_content("<a href='popup.html' target=_blank>Click me</a>")

    promise = asyncio.create_task(page.click("a"))
    [dialog, popup] = await asyncio.gather(
        page.context.wait_for_event("dialog"),
        page.wait_for_event("popup"),
    )

    assert dialog.message == "hey?"
    assert dialog.page == popup
    await dialog.accept("hello")
    await promise
    assert await popup.evaluate("window.result") == "hello"


async def test_console_event_should_work_with_context_manager(page: Page) -> None:
    async with page.context.expect_console_message() as cm_info:
        await page.evaluate("() => console.log('hello')")
    message = await cm_info.value
    assert message.text == "hello"
    assert message.page == page


async def test_page_error_event_should_work(page: Page) -> None:
    async with page.context.expect_event("weberror") as page_error_info:
        await page.set_content('<script>throw new Error("boom")</script>')
    page_error = await page_error_info.value
    assert page_error.page == page
    assert "boom" in page_error.error.stack


async def test_weberror_event_should_work(context: BrowserContext, page: Page) -> None:
    async with context.expect_event("weberror") as error_info:
        await page.goto('data:text/html,<script>throw new Error("Test")</script>')
    error = await error_info.value
    assert error.page == page
    assert error.error.message == "Test"



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import abc
import asyncio
import contextlib
import gzip
import mimetypes
import pathlib
import socket
import threading
from contextlib import closing
from http import HTTPStatus
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)
from urllib.parse import urlparse

from autobahn.twisted.resource import WebSocketResource
from autobahn.twisted.websocket import WebSocketServerFactory, WebSocketServerProtocol
from OpenSSL import crypto
from pyee import EventEmitter
from twisted.internet import reactor as _twisted_reactor
from twisted.internet import ssl
from twisted.internet.selectreactor import SelectReactor
from twisted.web import http

from playwright._impl._path_utils import get_file_dirname

_dirname = get_file_dirname()
reactor = cast(SelectReactor, _twisted_reactor)


def find_free_port() -> int:
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


T = TypeVar("T")


class ExpectResponse(Generic[T]):
    def __init__(self) -> None:
        self._value: T

    @property
    def value(self) -> T:
        if not hasattr(self, "_value"):
            raise ValueError("no received value")
        return self._value


class TestServerRequest(http.Request):
    __test__ = False
    channel: "TestServerHTTPChannel"
    post_body: Optional[bytes] = None

    def process(self) -> None:
        server = self.channel.factory.server_instance
        if self.content:
            self.post_body = self.content.read()
            self.content.seek(0, 0)
        else:
            self.post_body = None
        path = urlparse(self.uri.decode()).path

        request_subscriber = server.request_subscribers.get(path)
        if request_subscriber:
            request_subscriber._loop.call_soon_threadsafe(
                request_subscriber.set_result, self
            )
            server.request_subscribers.pop(path)

        if path == "/ws":
            server._ws_resource.render(self)
            return

        if server.auth.get(path):
            authorization_header = self.requestHeaders.getRawHeaders("authorization")
            creds_correct = False
            if authorization_header:
                creds_correct = server.auth.get(path) == (
                    self.getUser().decode(),
                    self.getPassword().decode(),
                )
            if not creds_correct:
                self.setHeader(b"www-authenticate", 'Basic realm="Secure Area"')
                self.setResponseCode(HTTPStatus.UNAUTHORIZED)
                self.write(b"HTTP Error 401 Unauthorized: Access is denied")
                self.finish()
                return
        if server.csp.get(path):
            self.setHeader(b"Content-Security-Policy", server.csp[path])
        if server.routes.get(path):
            server.routes[path](self)
            return

        self._serve_file(server.static_path / path[1:], path)

    def serve_file(self, path: pathlib.Path) -> None:
        return self._serve_file(path, urlparse(self.uri.decode()).path)

    def _serve_file(self, path: pathlib.Path, request_path: str) -> None:
        server = self.channel.factory.server_instance
        file_content = None
        try:
            file_content = path.read_bytes()
            content_type = mimetypes.guess_type(path)[0]
            if content_type and content_type.startswith("text/"):
                content_type += "; charset=utf-8"
            self.setHeader(b"Content-Type", content_type)
            self.setHeader(b"Cache-Control", "no-cache, no-store")
            if request_path in server.gzip_routes:
                self.setHeader("Content-Encoding", "gzip")
                self.write(gzip.compress(file_content))
            else:
                self.setHeader(b"Content-Length", str(len(file_content)))
                self.write(file_content)
            self.setResponseCode(HTTPStatus.OK)
        except (FileNotFoundError, IsADirectoryError, PermissionError):
            self.setHeader(b"Content-Type", "text/plain")
            self.setResponseCode(HTTPStatus.NOT_FOUND)
            if self.method != "HEAD":
                self.write(f"File not found: {path}".encode())
        self.finish()


class TestServerHTTPChannel(http.HTTPChannel):
    factory: "TestServerFactory"
    requestFactory = TestServerRequest


class TestServerFactory(http.HTTPFactory):
    server_instance: "Server"
    protocol = TestServerHTTPChannel


class Server:
    protocol = "http"

    def __init__(self) -> None:
        self.PORT = find_free_port()
        self.EMPTY_PAGE = f"{self.protocol}://localhost:{self.PORT}/empty.html"
        self.PREFIX = f"{self.protocol}://localhost:{self.PORT}"
        self.CROSS_PROCESS_PREFIX = f"{self.protocol}://127.0.0.1:{self.PORT}"
        # On Windows, this list can be empty, reporting text/plain for scripts.
        mimetypes.add_type("text/html", ".html")
        mimetypes.add_type("text/css", ".css")
        mimetypes.add_type("application/javascript", ".js")
        mimetypes.add_type("image/png", ".png")
        mimetypes.add_type("font/woff2", ".woff2")

    def __repr__(self) -> str:
        return self.PREFIX

    @abc.abstractmethod
    def listen(self, factory: TestServerFactory) -> None:
        pass

    def start(self, static_path: pathlib.Path = _dirname / "assets") -> None:
        request_subscribers: Dict[str, asyncio.Future] = {}
        auth: Dict[str, Tuple[str, str]] = {}
        csp: Dict[str, str] = {}
        routes: Dict[str, Callable[[TestServerRequest], Any]] = {}
        gzip_routes: Set[str] = set()
        self.request_subscribers = request_subscribers
        self.auth = auth
        self.csp = csp
        self.routes = routes
        self._ws_handlers: List[Callable[["WebSocketProtocol"], None]] = []
        self.gzip_routes = gzip_routes
        self.static_path = static_path
        factory = TestServerFactory()
        factory.server_instance = self

        ws_factory = WebSocketServerFactory()
        ws_factory.protocol = WebSocketProtocol
        setattr(ws_factory, "server_instance", self)
        self._ws_resource = WebSocketResource(ws_factory)

        self.listen(factory)

    async def wait_for_request(self, path: str) -> TestServerRequest:
        if path in self.request_subscribers:
            return await self.request_subscribers[path]
        future: asyncio.Future["TestServerRequest"] = asyncio.Future()
        self.request_subscribers[path] = future
        return await future

    def wait_for_web_socket(self) -> 'asyncio.Future["WebSocketProtocol"]':
        future: asyncio.Future[WebSocketProtocol] = asyncio.Future()
        self.once_web_socket_connection(future.set_result)
        return future

    @contextlib.contextmanager
    def expect_request(
        self, path: str
    ) -> Generator[ExpectResponse[TestServerRequest], None, None]:
        future = asyncio.create_task(self.wait_for_request(path))

        cb_wrapper: ExpectResponse[TestServerRequest] = ExpectResponse()

        def done_cb(task: asyncio.Task) -> None:
            cb_wrapper._value = future.result()

        future.add_done_callback(done_cb)
        yield cb_wrapper

    @contextlib.contextmanager
    def expect_websocket(
        self,
    ) -> Generator[ExpectResponse["WebSocketProtocol"], None, None]:
        future = self.wait_for_web_socket()

        cb_wrapper: ExpectResponse["WebSocketProtocol"] = ExpectResponse()

        def done_cb(_: asyncio.Future) -> None:
            cb_wrapper._value = future.result()

        future.add_done_callback(done_cb)
        yield cb_wrapper

    def set_auth(self, path: str, username: str, password: str) -> None:
        self.auth[path] = (username, password)

    def set_csp(self, path: str, value: str) -> None:
        self.csp[path] = value

    def reset(self) -> None:
        self.request_subscribers.clear()
        self.auth.clear()
        self.csp.clear()
        self.gzip_routes.clear()
        self.routes.clear()
        self._ws_handlers.clear()

    def set_route(
        self, path: str, callback: Callable[[TestServerRequest], Any]
    ) -> None:
        self.routes[path] = callback

    def enable_gzip(self, path: str) -> None:
        self.gzip_routes.add(path)

    def set_redirect(self, from_: str, to: str) -> None:
        def handle_redirect(request: http.Request) -> None:
            request.setResponseCode(HTTPStatus.FOUND)
            request.setHeader("location", to)
            request.finish()

        self.set_route(from_, handle_redirect)

    def send_on_web_socket_connection(self, data: bytes) -> None:
        self.once_web_socket_connection(lambda ws: ws.sendMessage(data))

    def once_web_socket_connection(
        self, handler: Callable[["WebSocketProtocol"], None]
    ) -> None:
        self._ws_handlers.append(handler)


class HTTPServer(Server):
    def __init__(self) -> None:
        self._listeners: list[Any] = []
        super().__init__()

    def listen(self, factory: http.HTTPFactory) -> None:
        self._listeners.append(
            reactor.listenTCP(self.PORT, factory, interface="127.0.0.1")
        )
        try:
            self._listeners.append(
                reactor.listenTCP(self.PORT, factory, interface="::1")
            )
        except Exception:
            pass

    def stop(self) -> None:
        for listener in self._listeners:
            listener.stopListening()
        self._listeners.clear()


class HTTPSServer(Server):
    protocol = "https"

    def __init__(self) -> None:
        self._listeners: list[Any] = []
        super().__init__()

    def listen(self, factory: http.HTTPFactory) -> None:
        cert = ssl.PrivateCertificate.fromCertificateAndKeyPair(
            ssl.Certificate.loadPEM(
                (_dirname / "testserver" / "cert.pem").read_bytes()
            ),
            ssl.KeyPair.load(
                (_dirname / "testserver" / "key.pem").read_bytes(), crypto.FILETYPE_PEM
            ),
        )
        contextFactory = cert.options()
        self._listeners.append(
            reactor.listenSSL(self.PORT, factory, contextFactory, interface="127.0.0.1")
        )
        try:
            self._listeners.append(
                reactor.listenSSL(self.PORT, factory, contextFactory, interface="::1")
            )
        except Exception:
            pass

    def stop(self) -> None:
        for listener in self._listeners:
            listener.stopListening()
        self._listeners.clear()


class WebSocketProtocol(WebSocketServerProtocol):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.events = EventEmitter()

    def onClose(self, wasClean: bool, code: int, reason: str) -> None:
        super().onClose(wasClean, code, reason)
        self.events.emit(
            "close",
            code,
            reason,
        )

    def onMessage(self, payload: Union[str, bytes], isBinary: bool) -> None:
        self.events.emit("message", payload, isBinary)

    def onOpen(self) -> None:
        for handler in getattr(self.factory, "server_instance")._ws_handlers.copy():
            getattr(self.factory, "server_instance")._ws_handlers.remove(handler)
            handler(self)


class TestServer:
    def __init__(self) -> None:
        self.server = HTTPServer()
        self.https_server = HTTPSServer()

    def start(self) -> None:
        self.server.start()
        self.https_server.start()
        self.thread = threading.Thread(
            target=lambda: reactor.run(installSignalHandlers=False)
        )
        self.thread.start()

    def stop(self) -> None:
        reactor.stop()
        self.thread.join()

    def reset(self) -> None:
        self.server.reset()
        self.https_server.reset()


test_server = TestServer()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from playwright._impl._browser import Browser
from playwright._impl._errors import Error
from playwright._impl._selectors import Selectors
from playwright.async_api import Page


async def test_should_work_with_internal_and(page: Page) -> None:
    await page.set_content(
        """
        <div class=foo>hello</div><div class=bar>world</div>
        <span class=foo>hello2</span><span class=bar>world2</span>
    """
    )
    assert (
        await page.eval_on_selector_all(
            'div >> internal:and="span"', "els => els.map(e => e.textContent)"
        )
    ) == []
    assert (
        await page.eval_on_selector_all(
            'div >> internal:and=".foo"', "els => els.map(e => e.textContent)"
        )
    ) == ["hello"]
    assert (
        await page.eval_on_selector_all(
            'div >> internal:and=".bar"', "els => els.map(e => e.textContent)"
        )
    ) == ["world"]
    assert (
        await page.eval_on_selector_all(
            'span >> internal:and="span"', "els => els.map(e => e.textContent)"
        )
    ) == ["hello2", "world2"]
    assert (
        await page.eval_on_selector_all(
            '.foo >> internal:and="div"', "els => els.map(e => e.textContent)"
        )
    ) == ["hello"]
    assert (
        await page.eval_on_selector_all(
            '.bar >> internal:and="span"', "els => els.map(e => e.textContent)"
        )
    ) == ["world2"]


async def test_should_throw_already_registered_error_when_registering(
    selectors: Selectors,
    browser: Browser,
) -> None:
    create_tag_selector = """
    () => ({
        query(root, selector) {
            return root.querySelector(selector);
        },
        queryAll(root, selector) {
            return Array.from(root.querySelectorAll(selector));
        }
    })
    """
    name = f"alreadyRegistered-{browser.browser_type.name}"
    await selectors.register(name, create_tag_selector)
    with pytest.raises(
        Error,
        match=f'Selectors.register: "{name}" selector engine has been already registered',
    ):
        await selectors.register(name, create_tag_selector)



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
from collections import defaultdict
from typing import Any

import objgraph
import pytest

from playwright.async_api import async_playwright
from tests.server import Server


@pytest.mark.asyncio
async def test_memory_objects(server: Server, browser_name: str) -> None:
    async with async_playwright() as p:
        browser = await p[browser_name].launch()
        page = await browser.new_page()
        await page.goto(server.EMPTY_PAGE)

        page.on("dialog", lambda dialog: dialog.dismiss())
        for _ in range(100):
            await page.evaluate("""async () => alert()""")

        await page.route("**/*", lambda route, _: route.fulfill(body="OK"))

        def handle_network_response_received(event: Any) -> None:
            event["__pw__is_last_network_response_received_event"] = True

        if browser_name == "chromium":
            # https://github.com/microsoft/playwright-python/issues/1602
            client = await page.context.new_cdp_session(page)
            await client.send("Network.enable")

            client.on(
                "Network.responseReceived",
                handle_network_response_received,
            )

        for _ in range(100):
            response = await page.evaluate("""async () => (await fetch("/")).text()""")
            assert response == "OK"

        await browser.close()

    gc.collect()

    pw_objects: defaultdict = defaultdict(int)
    for o in objgraph.by_type("dict"):
        assert isinstance(o, dict)
        name = o.get("_type")
        # https://github.com/microsoft/playwright-python/issues/1602
        if o.get("__pw__is_last_network_response_received_event", False):
            assert False
        if not name:
            continue
        pw_objects[name] += 1

    assert "Dialog" not in pw_objects
    assert "Request" not in pw_objects
    assert "Route" not in pw_objects



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
from asyncio.futures import Future

import pytest

from playwright.async_api import Browser, ConsoleMessage, Error, Page, Worker
from tests.server import Server
from tests.utils import TARGET_CLOSED_ERROR_MESSAGE


async def test_workers_page_workers(page: Page, server: Server) -> None:
    async with page.expect_worker() as worker_info:
        await page.goto(server.PREFIX + "/worker/worker.html")
    worker = await worker_info.value
    assert "worker.js" in worker.url
    assert repr(worker) == f"<Worker url={worker.url!r}>"

    assert (
        await worker.evaluate('() => self["workerFunction"]()')
        == "worker function result"
    )

    await page.goto(server.EMPTY_PAGE)
    assert len(page.workers) == 0


async def test_workers_should_emit_created_and_destroyed_events(page: Page) -> None:
    worker_obj = None
    async with page.expect_event("worker") as event_info:
        worker_obj = await page.evaluate_handle(
            "() => new Worker(URL.createObjectURL(new Blob(['1'], {type: 'application/javascript'})))"
        )
    worker = await event_info.value
    worker_this_obj = await worker.evaluate_handle("() => this")
    worker_destroyed_promise: Future[Worker] = asyncio.Future()
    worker.once("close", lambda w: worker_destroyed_promise.set_result(w))
    await page.evaluate("workerObj => workerObj.terminate()", worker_obj)
    assert await worker_destroyed_promise == worker
    with pytest.raises(Error) as exc:
        await worker_this_obj.get_property("self")
    assert TARGET_CLOSED_ERROR_MESSAGE in exc.value.message


async def test_workers_should_report_console_logs(page: Page) -> None:
    async with page.expect_console_message() as message_info:
        await page.evaluate(
            '() => new Worker(URL.createObjectURL(new Blob(["console.log(1)"], {type: "application/javascript"})))'
        )
    message = await message_info.value
    assert message.text == "1"


async def test_workers_should_have_JSHandles_for_console_logs(
    page: Page, browser_name: str
) -> None:
    log_promise: "asyncio.Future[ConsoleMessage]" = asyncio.Future()
    page.on("console", lambda m: log_promise.set_result(m))
    await page.evaluate(
        "() => new Worker(URL.createObjectURL(new Blob(['console.log(1,2,3,this)'], {type: 'application/javascript'})))"
    )
    log = await log_promise
    if browser_name != "firefox":
        assert log.text == "1 2 3 DedicatedWorkerGlobalScope"
    else:
        assert log.text == "1 2 3 JSHandle@object"
    assert len(log.args) == 4
    assert await (await log.args[3].get_property("origin")).json_value() == "null"


async def test_workers_should_evaluate(page: Page) -> None:
    async with page.expect_event("worker") as event_info:
        await page.evaluate(
            "() => new Worker(URL.createObjectURL(new Blob(['console.log(1)'], {type: 'application/javascript'})))"
        )
    worker = await event_info.value
    assert await worker.evaluate("1+1") == 2


async def test_workers_should_report_errors(page: Page) -> None:
    error_promise: "asyncio.Future[Error]" = asyncio.Future()
    page.on("pageerror", lambda e: error_promise.set_result(e))
    await page.evaluate(
        """() => new Worker(URL.createObjectURL(new Blob([`
      setTimeout(() => {
        // Do a console.log just to check that we do not confuse it with an error.
        console.log('hey');
        throw new Error('this is my error');
      })
    `], {type: 'application/javascript'})))"""
    )
    error_log = await error_promise
    assert "this is my error" in error_log.message


async def test_workers_should_clear_upon_navigation(server: Server, page: Page) -> None:
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_event("worker") as event_info:
        await page.evaluate(
            '() => new Worker(URL.createObjectURL(new Blob(["console.log(1)"], {type: "application/javascript"})))'
        )
    worker = await event_info.value
    assert len(page.workers) == 1
    destroyed = []
    worker.once("close", lambda _: destroyed.append(True))
    await page.goto(server.PREFIX + "/one-style.html")
    assert destroyed == [True]
    assert len(page.workers) == 0


async def test_workers_should_clear_upon_cross_process_navigation(
    server: Server, page: Page
) -> None:
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_event("worker") as event_info:
        await page.evaluate(
            "() => new Worker(URL.createObjectURL(new Blob(['console.log(1)'], {type: 'application/javascript'})))"
        )
    worker = await event_info.value
    assert len(page.workers) == 1
    destroyed = []
    worker.once("close", lambda _: destroyed.append(True))
    await page.goto(server.CROSS_PROCESS_PREFIX + "/empty.html")
    assert destroyed == [True]
    assert len(page.workers) == 0


@pytest.mark.skip_browser(
    "firefox"
)  # https://github.com/microsoft/playwright/issues/21760
async def test_workers_should_report_network_activity(
    page: Page, server: Server
) -> None:
    async with page.expect_worker() as worker_info:
        await page.goto(server.PREFIX + "/worker/worker.html")
    worker = await worker_info.value
    url = server.PREFIX + "/one-style.css"
    async with (
        page.expect_request(url) as request_info,
        page.expect_response(url) as response_info,
    ):
        await worker.evaluate(
            "url => fetch(url).then(response => response.text()).then(console.log)", url
        )
    request = await request_info.value
    response = await response_info.value
    assert request.url == url
    assert response.request == request
    assert response.ok


@pytest.mark.skip_browser(
    "firefox"
)  # https://github.com/microsoft/playwright/issues/21760
async def test_workers_should_report_network_activity_on_worker_creation(
    page: Page, server: Server
) -> None:
    # Chromium needs waitForDebugger enabled for this one.
    await page.goto(server.EMPTY_PAGE)
    url = server.PREFIX + "/one-style.css"
    async with (
        page.expect_request(url) as request_info,
        page.expect_response(url) as response_info,
    ):
        await page.evaluate(
            """url => new Worker(URL.createObjectURL(new Blob([`
        fetch("${url}").then(response => response.text()).then(console.log);
        `], {type: 'application/javascript'})))""",
            url,
        )
    request = await request_info.value
    response = await response_info.value
    assert request.url == url
    assert response.request == request
    assert response.ok


async def test_workers_should_format_number_using_context_locale(
    browser: Browser, server: Server
) -> None:
    context = await browser.new_context(locale="ru-RU")
    page = await context.new_page()
    await page.goto(server.EMPTY_PAGE)
    async with page.expect_worker() as worker_info:
        await page.evaluate(
            "() => new Worker(URL.createObjectURL(new Blob(['console.log(1)'], {type: 'application/javascript'})))"
        )
    worker = await worker_info.value
    assert await worker.evaluate("() => (10000.20).toLocaleString()") == "10\u00a0000,2"
    await context.close()



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

import pytest

from playwright.async_api import BrowserContext, Page
from tests.server import Server


async def test_should_return_no_cookies_in_pristine_browser_context(
    context: BrowserContext,
) -> None:
    assert await context.cookies() == []


async def test_should_get_a_cookie(
    context: BrowserContext,
    page: Page,
    server: Server,
    default_same_site_cookie_value: str,
) -> None:
    await page.goto(server.EMPTY_PAGE)
    document_cookie = await page.evaluate(
        """() => {
    document.cookie = 'username=John Doe';
    return document.cookie;
  }"""
    )
    assert document_cookie == "username=John Doe"
    assert await context.cookies() == [
        {
            "name": "username",
            "value": "John Doe",
            "domain": "localhost",
            "path": "/",
            "expires": -1,
            "httpOnly": False,
            "secure": False,
            "sameSite": default_same_site_cookie_value,
        }
    ]


async def test_should_get_a_non_session_cookie(
    context: BrowserContext,
    page: Page,
    server: Server,
    default_same_site_cookie_value: str,
) -> None:
    await page.goto(server.EMPTY_PAGE)
    # @see https://en.wikipedia.org/wiki/Year_2038_problem
    date = int(datetime.datetime(2038, 1, 1).timestamp() * 1000)
    document_cookie = await page.evaluate(
        """timestamp => {
    const date = new Date(timestamp);
    document.cookie = `username=John Doe;expires=${date.toUTCString()}`;
    return document.cookie;
  }""",
        date,
    )
    assert document_cookie == "username=John Doe"
    cookies = await context.cookies()
    expires = cookies[0]["expires"]
    del cookies[0]["expires"]
    # Browsers start to cap cookies with 400 days max expires value.
    # See https://github.com/httpwg/http-extensions/pull/1732
    # Chromium patch: https://chromium.googlesource.com/chromium/src/+/aaa5d2b55478eac2ee642653dcd77a50ac3faff6
    # We want to make sure that expires date is at least 400 days in future.
    # We use 355 to prevent flakes and not think about timezones!
    assert datetime.datetime.fromtimestamp(
        expires
    ) - datetime.datetime.now() > datetime.timedelta(days=355)
    assert cookies == [
        {
            "name": "username",
            "value": "John Doe",
            "domain": "localhost",
            "path": "/",
            "httpOnly": False,
            "secure": False,
            "sameSite": default_same_site_cookie_value,
        }
    ]


async def test_should_properly_report_httpOnly_cookie(
    context: BrowserContext, page: Page, server: Server
) -> None:
    server.set_route(
        "/empty.html",
        lambda r: (
            r.setHeader("Set-Cookie", "name=value;httpOnly; Path=/"),
            r.finish(),
        ),
    )

    await page.goto(server.EMPTY_PAGE)
    cookies = await context.cookies()
    assert len(cookies) == 1
    assert cookies[0]["httpOnly"] is True


async def test_should_properly_report_strict_sameSite_cookie(
    context: BrowserContext, page: Page, server: Server, is_webkit: bool, is_win: bool
) -> None:
    if is_webkit and is_win:
        pytest.skip()

    server.set_route(
        "/empty.html",
        lambda r: (
            r.setHeader("Set-Cookie", "name=value;sameSite=Strict"),
            r.finish(),
        ),
    )
    await page.goto(server.EMPTY_PAGE)
    cookies = await context.cookies()
    assert len(cookies) == 1
    assert cookies[0]["sameSite"] == "Strict"


async def test_should_properly_report_lax_sameSite_cookie(
    context: BrowserContext, page: Page, server: Server, is_webkit: bool, is_win: bool
) -> None:
    if is_webkit and is_win:
        pytest.skip()

    server.set_route(
        "/empty.html",
        lambda r: (
            r.setHeader("Set-Cookie", "name=value;sameSite=Lax"),
            r.finish(),
        ),
    )
    await page.goto(server.EMPTY_PAGE)
    cookies = await context.cookies()
    assert len(cookies) == 1
    assert cookies[0]["sameSite"] == "Lax"


async def test_should_get_multiple_cookies(
    context: BrowserContext,
    page: Page,
    server: Server,
    default_same_site_cookie_value: str,
) -> None:
    await page.goto(server.EMPTY_PAGE)
    document_cookie = await page.evaluate(
        """() => {
    document.cookie = 'username=John Doe';
    document.cookie = 'password=1234';
    return document.cookie.split('; ').sort().join('; ');
  }"""
    )
    cookies = await context.cookies()
    cookies.sort(key=lambda r: r["name"])
    assert document_cookie == "password=1234; username=John Doe"
    assert cookies == [
        {
            "name": "password",
            "value": "1234",
            "domain": "localhost",
            "path": "/",
            "expires": -1,
            "httpOnly": False,
            "secure": False,
            "sameSite": default_same_site_cookie_value,
        },
        {
            "name": "username",
            "value": "John Doe",
            "domain": "localhost",
            "path": "/",
            "expires": -1,
            "httpOnly": False,
            "secure": False,
            "sameSite": default_same_site_cookie_value,
        },
    ]


async def test_should_get_cookies_from_multiple_urls(
    context: BrowserContext, default_same_site_cookie_value: str
) -> None:
    await context.add_cookies(
        [
            {"url": "https://foo.com", "name": "doggo", "value": "woofs"},
            {"url": "https://bar.com", "name": "catto", "value": "purrs"},
            {"url": "https://baz.com", "name": "birdo", "value": "tweets"},
        ]
    )
    cookies = await context.cookies(["https://foo.com", "https://baz.com"])
    cookies.sort(key=lambda r: r["name"])

    assert cookies == [
        {
            "name": "birdo",
            "value": "tweets",
            "domain": "baz.com",
            "path": "/",
            "expires": -1,
            "httpOnly": False,
            "secure": True,
            "sameSite": default_same_site_cookie_value,
        },
        {
            "name": "doggo",
            "value": "woofs",
            "domain": "foo.com",
            "path": "/",
            "expires": -1,
            "httpOnly": False,
            "secure": True,
            "sameSite": default_same_site_cookie_value,
        },
    ]



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict

import pytest

from playwright.sync_api import BrowserContext, BrowserType


def test_context_managers(browser_type: BrowserType, launch_arguments: Dict) -> None:
    with browser_type.launch(**launch_arguments) as browser:
        with browser.new_context() as context:
            with context.new_page():
                assert len(context.pages) == 1
            assert len(context.pages) == 0
            assert len(browser.contexts) == 1
        assert len(browser.contexts) == 0
    assert not browser.is_connected()


def test_context_managers_not_hang(context: BrowserContext) -> None:
    with pytest.raises(Exception, match="Oops!"):
        with context.new_page():
            raise Exception("Oops!")



# Copyright (c) Microsoft Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import pytest

from playwright.sync_api import Error, Locator, Page, expect
from tests.server import Server
from tests.utils import TARGET_CLOSED_ERROR_MESSAGE


def test_should_work(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")

    before_count = 0
    after_count = 0

    original_locator = page.get_by_text("This interstitial covers the button")

    def handler(locator: Locator) -> None:
        assert locator == original_locator
        nonlocal before_count
        nonlocal after_count
        before_count += 1
        page.locator("#close").click()
        after_count += 1

    page.add_locator_handler(original_locator, handler)

    for args in [
        ["mouseover", 1],
        ["mouseover", 1, "capture"],
        ["mouseover", 2],
        ["mouseover", 2, "capture"],
        ["pointerover", 1],
        ["pointerover", 1, "capture"],
        ["none", 1],
        ["remove", 1],
        ["hide", 1],
    ]:
        page.locator("#aside").hover()
        before_count = 0
        after_count = 0
        page.evaluate(
            "(args) => { window.clicked = 0; window.setupAnnoyingInterstitial(...args); }",
            args,
        )
        assert before_count == 0
        assert after_count == 0
        page.locator("#target").click()
        assert before_count == args[1]
        assert after_count == args[1]
        assert page.evaluate("window.clicked") == 1
        expect(page.locator("#interstitial")).not_to_be_visible()


def test_should_work_with_a_custom_check(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")

    def handler() -> None:
        if page.get_by_text("This interstitial covers the button").is_visible():
            page.locator("#close").click()

    page.add_locator_handler(page.locator("body"), handler, no_wait_after=True)

    for args in [
        ["mouseover", 2],
        ["none", 1],
        ["remove", 1],
        ["hide", 1],
    ]:
        page.locator("#aside").hover()
        page.evaluate(
            "(args) => { window.clicked = 0; window.setupAnnoyingInterstitial(...args); }",
            args,
        )
        page.locator("#target").click()
        assert page.evaluate("window.clicked") == 1
        expect(page.locator("#interstitial")).not_to_be_visible()


def test_should_work_with_locator_hover(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")

    page.add_locator_handler(
        page.get_by_text("This interstitial covers the button"),
        lambda: page.locator("#close").click(),
    )

    page.locator("#aside").hover()
    page.evaluate(
        '() => { window.setupAnnoyingInterstitial("pointerover", 1, "capture"); }'
    )
    page.locator("#target").hover()
    expect(page.locator("#interstitial")).not_to_be_visible()
    assert (
        page.eval_on_selector(
            "#target", "e => window.getComputedStyle(e).backgroundColor"
        )
        == "rgb(255, 255, 0)"
    )


def test_should_not_work_with_force_true(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")

    page.add_locator_handler(
        page.get_by_text("This interstitial covers the button"),
        lambda: page.locator("#close").click(),
    )

    page.locator("#aside").hover()
    page.evaluate('() => { window.setupAnnoyingInterstitial("none", 1); }')
    page.locator("#target").click(force=True, timeout=2000)
    assert page.locator("#interstitial").is_visible()
    assert page.evaluate("window.clicked") is None


def test_should_throw_when_page_closes(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")

    page.add_locator_handler(
        page.get_by_text("This interstitial covers the button"), lambda: page.close()
    )

    page.locator("#aside").hover()
    page.evaluate(
        '() => { window.clicked = 0; window.setupAnnoyingInterstitial("mouseover", 1); }'
    )
    with pytest.raises(Error) as exc:
        page.locator("#target").click()
    assert TARGET_CLOSED_ERROR_MESSAGE in exc.value.message


def test_should_throw_when_handler_times_out(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")

    called = 0

    def handler() -> None:
        nonlocal called
        called += 1
        # Deliberately timeout.
        try:
            page.wait_for_timeout(9999999)
        except Exception:
            pass

    page.add_locator_handler(
        page.get_by_text("This interstitial covers the button"), handler
    )

    page.locator("#aside").hover()
    page.evaluate(
        '() => { window.clicked = 0; window.setupAnnoyingInterstitial("mouseover", 1); }'
    )
    with pytest.raises(Error) as exc:
        page.locator("#target").click(timeout=3000)
    assert "Timeout 3000ms exceeded" in exc.value.message

    with pytest.raises(Error) as exc:
        page.locator("#target").click(timeout=3000)
    assert "Timeout 3000ms exceeded" in exc.value.message

    # Should not enter the same handler while it is still running.
    assert called == 1


def test_should_work_with_to_be_visible(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")

    called = 0

    def handler() -> None:
        nonlocal called
        called += 1
        page.locator("#close").click()

    page.add_locator_handler(
        page.get_by_text("This interstitial covers the button"), handler
    )

    page.evaluate(
        '() => { window.clicked = 0; window.setupAnnoyingInterstitial("remove", 1); }'
    )
    expect(page.locator("#target")).to_be_visible()
    expect(page.locator("#interstitial")).not_to_be_visible()
    assert called == 1


def test_should_work_when_owner_frame_detaches(page: Page, server: Server) -> None:
    page.goto(server.EMPTY_PAGE)
    page.evaluate(
        """
    () => {
        const iframe = document.createElement('iframe');
        iframe.src = 'data:text/html,<body>hello from iframe</body>';
        document.body.append(iframe);

        const target = document.createElement('button');
        target.textContent = 'Click me';
        target.id = 'target';
        target.addEventListener('click', () => window._clicked = true);
        document.body.appendChild(target);

        const closeButton = document.createElement('button');
        closeButton.textContent = 'close';
        closeButton.id = 'close';
        closeButton.addEventListener('click', () => iframe.remove());
        document.body.appendChild(closeButton);
    }
    """
    )
    page.add_locator_handler(
        page.frame_locator("iframe").locator("body"),
        lambda: page.locator("#close").click(),
    )
    page.locator("#target").click()
    assert page.query_selector("iframe") is None
    assert page.evaluate("window._clicked") is True


def test_should_work_with_times_option(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")
    called = 0

    def _handler() -> None:
        nonlocal called
        called += 1

    page.add_locator_handler(
        page.locator("body"), _handler, no_wait_after=True, times=2
    )
    page.locator("#aside").hover()
    page.evaluate(
        """
    () => {
        window.clicked = 0;
        window.setupAnnoyingInterstitial('mouseover', 4);
    }
    """
    )
    with pytest.raises(Error) as exc_info:
        page.locator("#target").click(timeout=3000)
    assert called == 2
    assert page.evaluate("window.clicked") == 0
    expect(page.locator("#interstitial")).to_be_visible()
    assert "Timeout 3000ms exceeded" in exc_info.value.message
    assert (
        '<div>This interstitial covers the button</div> from <div class="visible" id="interstitial">…</div> subtree intercepts pointer events'
        in exc_info.value.message
    )


def test_should_wait_for_hidden_by_default(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")
    called = 0

    def _handler(button: Locator) -> None:
        nonlocal called
        called += 1
        button.click()

    page.add_locator_handler(page.get_by_role("button", name="close"), _handler)
    page.locator("#aside").hover()
    page.evaluate(
        """
    () => {
        window.clicked = 0;
        window.setupAnnoyingInterstitial('timeout', 1);
    }
    """
    )
    page.locator("#target").click()
    assert page.evaluate("window.clicked") == 1
    expect(page.locator("#interstitial")).not_to_be_visible()
    assert called == 1


def test_should_wait_for_hidden_by_default_2(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")
    called = 0

    def _handler() -> None:
        nonlocal called
        called += 1

    page.add_locator_handler(page.get_by_role("button", name="close"), _handler)
    page.locator("#aside").hover()
    page.evaluate(
        """
    () => {
        window.clicked = 0;
        window.setupAnnoyingInterstitial('hide', 1);
    }
    """
    )
    with pytest.raises(Error) as exc_info:
        page.locator("#target").click(timeout=3000)
    assert page.evaluate("window.clicked") == 0
    assert page.locator("#interstitial").is_visible()
    assert called == 1
    assert (
        'locator handler has finished, waiting for get_by_role("button", name="close") to be hidden'
        in exc_info.value.message
    )


def test_should_work_with_noWaitAfter(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")
    called = 0

    def _handler(button: Locator) -> None:
        nonlocal called
        called += 1
        if called == 1:
            button.click()
        else:
            page.locator("#interstitial").wait_for(state="hidden")

    page.add_locator_handler(
        page.get_by_role("button", name="close"), _handler, no_wait_after=True
    )
    page.locator("#aside").hover()
    page.evaluate(
        """
    () => {
        window.clicked = 0;
        window.setupAnnoyingInterstitial('timeout', 1);
    }
    """
    )
    page.locator("#target").click()
    assert page.evaluate("window.clicked") == 1
    expect(page.locator("#interstitial")).not_to_be_visible()
    assert called == 2


def test_should_removeLocatorHandler(page: Page, server: Server) -> None:
    page.goto(server.PREFIX + "/input/handle-locator.html")
    called = 0

    def _handler(locator: Locator) -> None:
        nonlocal called
        called += 1
        locator.click()

    page.add_locator_handler(page.get_by_role("button", name="close"), _handler)
    page.evaluate(
        """
    () => {
        window.clicked = 0;
        window.setupAnnoyingInterstitial('hide', 1);
    }
    """
    )
    page.locator("#target").click()
    assert called == 1
    assert page.evaluate("window.clicked") == 1
    expect(page.locator("#interstitial")).not_to_be_visible()
    page.evaluate(
        """
    () => {
        window.clicked = 0;
        window.setupAnnoyingInterstitial('hide', 1);
    }
    """
    )
    page.remove_locator_handler(page.get_by_role("button", name="close"))
    with pytest.raises(Error) as error:
        page.locator("#target").click(timeout=3000)
    assert called == 1
    assert page.evaluate("window.clicked") == 0
    expect(page.locator("#interstitial")).to_be_visible()
    assert "Timeout 3000ms exceeded" in error.value.message
